import os
import tempfile
from typing import Optional, Sequence, List, Union
from osgeo import gdal, ogr, osr
from gdalos.rectangle import GeoRectangle
from gdalos.calc import gdal_to_czml
from gdalos.gdalos_color import ColorPalette, get_file_from_strings
from gdalos import gdalos_util
from gdalos.calc.gdalos_raster_color import DiscreteMode, gdalos_raster_color
import copy


# def get_named_temporary_filenme(suffix='', dir_name=''):
#     filename = next(tempfile._get_candidate_names())+suffix
#     if dir_name is None:
#         return filename
#     else:
#         if dir_name=='':
#             dir_name = tempfile._get_default_tempdir()
#         return os.path.join(dir_name, filename)


def gdal_crop(ds: gdal.Dataset, out_filename: str, output_format: str = 'MEM',
              extent: Optional[GeoRectangle] = None,
              cutline: Optional[Union[str, List[str]]] = None,
              common_options: dict = None):
    common_options = dict(common_options or dict())
    translate_options = {}
    warp_options = {}
    temp_filename = None

    if cutline is None:
        if extent is None:
            return ds
        # -projwin minx maxy maxx miny (ulx uly lrx lry)
        translate_options['projWin'] = extent.lurd
        # -te minx miny maxx maxy
    else:
        if extent is not None:
            warp_options['outputBounds'] = extent.ldru
        warp_options['cropToCutline'] = extent is None
        warp_options['dstNodata'] = -32768
        # warp_options['dstAlpha'] = True

        if isinstance(cutline, str):
            cutline_filename = cutline
        elif isinstance(cutline, Sequence):
            temp_filename = tempfile.mktemp(suffix='.gpkg')
            cutline_filename = temp_filename
            gdalos_util.wkt_write_ogr(cutline_filename, cutline, of='GPKG')
        warp_options['cutlineDSName'] = cutline_filename
    # else:
    #     raise Exception("extent is unknown {}".format(extent))

    common_options['format'] = output_format
    if warp_options:
        ds = gdal.Warp(str(out_filename), ds, **common_options, **warp_options)
    else:
        ds = gdal.Translate(str(out_filename), ds, **common_options, **translate_options)
    if temp_filename is not None:
        os.remove(temp_filename)
    return ds


def czml_gdaldem_crop_and_color(
        ds: gdal.Dataset,
        out_filename: str = None, output_format: str = None,
        czml_output_filename: str = None,
        extent: Optional[GeoRectangle] = None,
        cutline: Optional[Union[str, List[str]]] = None,
        color_palette: ColorPalette = None,
        discrete_mode=DiscreteMode.interp,
        process_palette=None,
        common_options: dict = None):
    ds = gdalos_crop(
        ds, out_filename=out_filename, output_format=output_format,
        extent=extent, cutline=cutline,
        common_options=common_options)

    min_max = gdalos_util.get_raster_min_max(ds) if process_palette and color_palette.has_percents else None

    # output_format_crop = 'MEM' if do_color else output_format
    ds = gdalos_raster_color(
        ds, color_palette=color_palette,
        out_filename=out_filename, output_format=output_format,
        discrete_mode=discrete_mode)

    if ds is None:
        raise Exception('fail to color')
    if czml_output_filename is not None:
        if min_max:
            color_palette_copy = copy.deepcopy(color_palette)
            color_palette_copy.apply_percent(*min_max)
        else:
            color_palette_copy = color_palette
        meta = gdal_to_czml.make_czml_description(color_palette_copy, process_palette)
        ds.SetMetadataItem(gdal_to_czml.czml_metadata_name, meta)

        gdal_to_czml.gdal_to_czml(ds, name=czml_output_filename, out_filename=czml_output_filename)
    return ds

def gdalos_crop(ds: gdal.Dataset,
                out_filename: str = '', output_format: str = None,
                extent: Optional[GeoRectangle] = None,
                cutline: Optional[Union[str, List[str]]] = None,
                common_options: dict = None):
    if (extent or cutline) is None:
        return ds

    if not output_format:
        output_format = 'GTiff' if out_filename else 'MEM'

    if not out_filename:
        out_filename = ''
        if output_format != 'MEM':
            raise Exception('output filename is None')

    ds = gdal_crop(ds, out_filename,
                   output_format=output_format, extent=extent, cutline=cutline,
                   common_options=common_options)
    if ds is None:
        raise Exception('fail to crop')

    return ds


# def gdaldem_crop_and_color(ds: gdal.Dataset,
#                            out_filename: str = None, output_format: str = None,
#                            extent: Optional[GeoRectangle] = None,
#                            cutline: Optional[Union[str, List[str]]] = None,
#                            color_palette: ColorPalette = None,
#                            discrete_mode=DiscreteMode.interp,
#                            get_min_max=None,
#                            common_options: dict = None):
#     ds = gdalos_crop(
#         ds, out_filename=out_filename, output_format=output_format,
#         extent=extent, cutline=cutline,
#         common_options=common_options)
#
#     min_max = get_raster_min_max(ds) if get_min_max else None
#
#     ds = gdalos_raster_color(
#         ds, color_palette=color_palette,
#         out_filename=out_filename, output_format=output_format,
#         discrete_mode=discrete_mode)
#
#     return ds, min_max


def get_wkt_list(filename):
    ds = ogr.Open(filename, 0)
    layer = ds.GetLayer()
    wkt_list = []
    for feature in layer:
        geom = feature.GetGeometryRef()
        wkt_list.append(geom.ExportToWkt())
    return wkt_list


def read_list(filename):
    return [line.rstrip('\n') for line in open(filename)]


if __name__ == '__main__':
    script_dir = os.path.dirname(os.path.realpath(__file__))
    root_data = os.path.join(script_dir, r'../../data/sample')
    shp_filename = os.path.join(root_data, r'shp/poly.shp')
    wkt_list = get_wkt_list(shp_filename)

    raster_filename = os.path.join(root_data, r'maps/srtm1_x35_y32.tif')
    ds = gdalos_util.open_ds(raster_filename)
    out_filename = tempfile.mktemp(suffix='.tif')

    ds = gdalos_crop(
        ds=ds,
        out_filename=out_filename,
        cutline=wkt_list,
        output_format='GTiff')

    print(out_filename)



from enum import Enum, auto


def enum_to_str(enum_or_str):
    return enum_or_str.name if isinstance(enum_or_str, Enum) else str(enum_or_str)


class OvrType(Enum):
    # existing_reuse or create_external_auto (by existance of src overviews)
    auto_select = auto()
    # work with existing overviews
    existing_reuse = auto()
    # create_external_single or create_external_multi (by size)
    create_external_auto = auto()
    # create a single .ovr file with all the overviews
    create_external_single = auto()
    # create one ovr file per overview: .ovr, .ovr.ovr, .ovr.ovr.orv ....
    create_external_multi = auto()
    # create overviews inside the main dataset file
    create_internal = auto()


# these are the common resamplers for translate, warp, addo, buildvrt.
# there are some specific resamplers for warp and addo
class GdalResamplingAlg(Enum):
    # nearest applies a nearest neighbour (simple sampling) resampler. for warp it's called 'near'
    nearest = auto()
    # average computes the average of all non-NODATA contributing pixels.
    # Starting with GDAL 3.1, this is a weighted average taking into account properly
    # the weight of source pixels not contributing fully to the target pixel.
    average = auto()
    # bilinear applies a bilinear convolution kernel.
    bilinear = auto()
    # cubic applies a cubic convolution kernel.
    cubic = auto()
    # cubicspline applies a B-Spline convolution kernel.
    cubicspline = auto()
    # lanczos applies a Lanczos windowed sinc convolution kernel.
    lanczos = auto()
    # mode selects the value which appears most often of all the sampled points.
    mode = auto()


class GdalOutputFormat(Enum):
    gtiff = auto()
    cog = auto()
    gpkg = auto()
    mem = auto()



from itertools import chain

from fidget.backend.QtWidgets import QFrame, QHBoxLayout
from fidget.widgets import (
    FidgetCheckBox,
    FidgetCombo,
    FidgetConverter,
    FidgetDict,
    FidgetEditCombo,
    FidgetFilePath,
    FidgetFloat,
    FidgetInt,
    FidgetMinimal,
    FidgetOptional,
    FidgetSpin,
    FidgetTabs,
    FidgetTuple,
    inner_fidget,
)
from gdalos.gdalos_main import GeoRectangle, OvrType, RasterKind, GdalResamplingAlg, GdalOutputFormat
from gdalos_qt.crs_widget import CrsWidget
from gdalos_qt.nodatavalue_widget import NodatavalueWidgetSrc, NodatavalueWidgetDst

raster_fidget = FidgetFilePath.template(
    dialog={
        "filter": "raster files (*.tif *.tiff *.vrt *.xml *.img *.dt1 *.dt2);;all files (*.*)"
    },
    make_title=True,
)

FidgetCheckBox.MAKE_TITLE = True
FidgetEditCombo.MAKE_INDICATOR = FidgetEditCombo.MAKE_PLAINTEXT = False


class GdalosWidget(FidgetConverter):
    @inner_fidget("gdalos")
    class GdalosWidget(FidgetTabs):
        INNER_TEMPLATES = [
            FidgetDict.template(
                "basic",
                [
                    (
                        "filename",
                        raster_fidget.template("source file", exist_cond=True),
                    ),
                    (
                        "cog",
                        FidgetCheckBox.template(
                            "make cog (cloud optimized geotiff)", initial_value=True
                        ),

                    ),
                    FidgetCheckBox.template("prefer_2_step_cog", initial_value=True),
                    FidgetCheckBox.template("delete_temp_files", initial_value=True),
                    FidgetCheckBox.template("overwrite", initial_value=False),
                    (
                        "multi_file_as_vrt",
                        FidgetCheckBox.template(
                            "treat multi file as vrt", initial_value=False
                        ),
                    ),
                    FidgetCheckBox.template("sparse_ok", initial_value=True),
                    FidgetCheckBox.template("write_info", initial_value=True),
                    FidgetCheckBox.template("write_spec", initial_value=True),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "overviews",
                [
                    (
                        "ovr_idx",
                        FidgetOptional.template(
                            FidgetInt.template("source overview index", placeholder=False),
                            make_title=True,
                        ),
                    ),
                    FidgetOptional.template(
                        FidgetInt.template("dst_ovr_count", placeholder=False),
                        make_title=True,
                    ),
                    FidgetCombo.template(
                        "ovr_type",
                        options=list(
                            chain((("AUTO", ...), ("None", None)), OvrType.__members__)
                        ),
                        initial_index=0,
                        make_title=True,
                    ),
                    FidgetCheckBox.template(
                        "keep_src_ovr_suffixes", initial_value=True
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "format",
                [
                    # todo validate drivers?
                    (
                        "of",
                        FidgetCombo.template(
                            "output format",
                            options=list(
                                chain((("AUTO", ...),), GdalOutputFormat.__members__)
                            ),
                            initial_index=0,
                            make_title=True,
                        ),
                    ),
                    (
                        "outext",
                        FidgetEditCombo.template(
                            "output extension", options=("tif",), make_title=True
                        ),
                    ),
                    FidgetCheckBox.template("tiled", options=("NO", "YES")),
                    FidgetCombo.template(
                        "big_tiff",
                        options=["YES", "NO", "IF_NEEDED", "IF_SAFER"],
                        initial_value="IF_SAFER",
                        make_title=True,
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "conversion",
                [
                    FidgetCombo.template(
                        "lossy",
                        options=[("AUTO", None), False, True],
                        initial_index=0,
                        make_title=True,
                    ),
                    FidgetCombo.template(
                        "kind",
                        options=list(
                            chain((("AUTO", ...),), RasterKind.__members__.items())
                        ),
                        initial_index=0,
                        make_title=True,
                    ),
                    FidgetCheckBox.template("expand_rgb", initial_value=False),
                    FidgetMinimal.template(
                        FidgetOptional.template(
                            CrsWidget.template(
                                "warp_srs",
                                make_plaintext=True,
                                make_indicator=True,
                                frame_style=QFrame.Box | QFrame.Plain,
                            )
                        ),
                        make_title=True,
                        make_plaintext=False,
                        make_indicator=False,
                        initial_value=None,
                    ),
                    (
                        "out_res",
                        FidgetOptional.template(
                            FidgetTuple.template(
                                "resolution",
                                inner_templates=(
                                    FidgetInt.template(
                                        "X", make_indicator=False, make_title=False
                                    ),
                                    FidgetInt.template(
                                        "Y", make_indicator=False, make_title=False
                                    ),
                                ),
                                layout_cls=QHBoxLayout,
                            ),
                            make_indicator=False,
                            make_title=True,
                            make_plaintext=True,
                        ),
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "extent",
                [
                    FidgetOptional.template(
                        FidgetConverter.template(
                            FidgetTuple.template(
                                "extent",
                                inner_templates=[
                                    FidgetFloat.template(letter) for letter in "WESN"
                                ],
                                layout_cls=QHBoxLayout,
                                make_title=False,
                            ),
                            converter_func=lambda x: GeoRectangle.from_lrdu(*x),
                            back_converter_func=GeoRectangle.lrdu.fget,
                        ),
                        make_title=True,
                        make_indicator=False,
                        make_plaintext=True,
                    ),
                    (
                        "extent_in_4326",
                        FidgetCheckBox.template(
                            "extent is in EPSG:4326 (WGS84Geo), otherwise it should be in target srs",
                            initial_value=True
                        ),
                    ),
                    (
                        "src_win",
                        FidgetOptional.template(
                            FidgetConverter.template(
                                FidgetTuple.template(
                                    "source window",
                                    inner_templates=[
                                        FidgetFloat.template(letter)
                                        for letter in "XYWH"
                                    ],
                                    layout_cls=QHBoxLayout,
                                    make_title=False,
                                ),
                                converter_func=lambda x: GeoRectangle(*x),
                                back_converter_func=GeoRectangle.xywh.fget,
                            ),
                            make_title=True,
                            make_indicator=False,
                            make_plaintext=True,
                        ),
                    ),
                    FidgetOptional.template(
                        FidgetInt.template("partition", placeholder=False),
                        make_title=True,
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                'NoData value',
                [
                    FidgetCheckBox.template("hide nodatavalue", initial_value=False),
                    (
                        "dst_nodatavalue",
                        NodatavalueWidgetDst.template("dst_nodatavalue"),
                    ),
                    (
                        "src_nodatavalue",
                        NodatavalueWidgetSrc.template("src_nodatavalue"),
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "output",
                [
                    (
                        "out_path",
                        FidgetOptional.template(
                            raster_fidget.template(
                                "output path: folder", exist_cond=None
                            ),
                            make_indicator=False,
                        ),
                    ),
                    (
                        "out_path_with_src_folders",
                        FidgetCheckBox.template(
                            "output path: keep source folder structure",
                            initial_value=True,
                        ),
                    ),
                    (
                        "out_filename",
                        FidgetOptional.template(
                            raster_fidget.template(
                                "output path: set destination file", exist_cond=None
                            ),
                            make_indicator=False,
                        ),
                    ),
                ],
                make_plaintext=False,
            ),
            FidgetDict.template(
                "advanced",
                [
                    FidgetCombo.template(
                        "resampling_alg",
                        options=list(
                            chain((("AUTO", ...), ("None", None)), GdalResamplingAlg.__members__)
                        ),
                        initial_index=0,
                        make_title=True,
                    ),
                    FidgetCheckBox.template("keep_alpha"),
                    FidgetSpin.template(
                        "quality", 1, 100, initial_value=75, make_title=True
                    ),
                ],
                make_plaintext=False,
            ),
        ]

    def convert(self, v: dict):
        ret = {}
        for k, v in v.items():
            ret.update(v)
        return ret

    def back_convert(self, v: dict):
        ret = {}
        sw: FidgetDict
        for d_name, sw in self.inner.inners.items():
            ret[d_name] = {}
            for title, ssw in sw.inners.items():
                if title in v:
                    ret[d_name][title] = v[title]

        return ret


if __name__ == "__main__":
    from fidget.backend.QtWidgets import QApplication

    app = QApplication([])
    w = GdalosWidget(make_title=True, make_plaintext=True, make_indicator=True)
    print(w.fill)
    w.show()
    res = app.exec_()
    print(w.value())
    exit(res)



def fill_arrays(*args):
    max_len = max(len(x) for x in args)
    result = []
    for x in range(len(args)):
        lx = len(x)
        if lx < max_len:
            v = x[lx-1]
            for i in range(lx, max_len):
                x.append(v)
        result.append(x)
    return result


def fill_arrays_dict(d: dict):
    max_len = max(len(v) for v in d.values())
    result = dict()
    for k, v in d.items():
        len_v = len(v)
        if len_v < max_len:
            v = v[len_v-1]
            for i in range(len_v, max_len):
                v.append(v)
        result[k] = v
    return result


def make_dicts_list_from_lists_dict(d: dict, key_map):
    max_len = max(len(v) for v in d.values())
    result = []
    d1 = dict()
    for i in range(max_len):
        d1 = d1.copy()
        for k, v in d.items():  # zip(new_keys, d.values()):
            if key_map:
                k = key_map[k]
            len_v = len(v)
            if i < len_v:
                d1[k] = v[i]
        result.append(d1)
    return result


def replace_keys(dicts, key_map):
    if not key_map:
        return dicts
    result = []
    single = isinstance(dicts, dict)
    if single:
        dicts = [dicts]
    for d in dicts:
        new_d = dict()
        for k, v in d.items():
            if key_map:
                k = key_map[k]
            new_d[k] = v
        result.append(new_d)
    if single:
        result = result[0]
    return result



from typing import Union, Optional

from fidget.backend.QtWidgets import QHBoxLayout
from fidget.widgets import FidgetConst, FidgetInt, FidgetStacked


class NodatavalueWidgetSrc(FidgetStacked[Optional[Union[float, type(...)]]]):
    INNER_TEMPLATES = [
        FidgetConst.template("get original value", option=("...", ...)),
        FidgetConst.template("raster minimum (might take time)", option=("None", None)),
        FidgetInt.template("given value"),
    ]
    SELECTOR_CLS = "radio"
    MAKE_TITLE = True
    MAKE_INDICATOR = False
    MAKE_PLAINTEXT = False
    LAYOUT_CLS = QHBoxLayout


class NodatavalueWidgetDst(FidgetStacked[Optional[Union[float, type(...)]]]):
    INNER_TEMPLATES = [
        FidgetConst.template("don't change", option=("None", None)),
        FidgetConst.template("change to default value (for DTM)", option=("...", ...)),
        FidgetInt.template("change to a set value"),
    ]
    SELECTOR_CLS = "radio"
    MAKE_TITLE = True
    MAKE_INDICATOR = False
    MAKE_PLAINTEXT = False
    LAYOUT_CLS = QHBoxLayout


if __name__ == "__main__":
    from fidget.backend.QtWidgets import QApplication
    from fidget.widgets import FidgetOptional

    app = QApplication([])
    # w = FidgetOptional(NodatavalueWidget("sample nodatavalue"))
    # w = NodatavalueWidgetSrc("sample nodatavalue", initial_value=None)  # todo: setting initial_value doesn't work
    w = NodatavalueWidgetSrc("sample nodatavalue")
    w.show()
    app.exec_()
    print(w.value())



import os
import re
import gdal
from xml.dom import minidom
from collections import OrderedDict
from pathlib import Path
import glob
import tempfile
from typing import Sequence, Union
import math


def to_number(s):
    try:
        return int(s)
    except ValueError:
        return float(s)


def byte(number, i):
    return (number & (0xff << (i * 8))) >> (i * 8)


PathOrStrings = Union[Path, str, Sequence[str]]
ColorPaletteOrPathOrStrings = Union['ColorPalette', PathOrStrings]


class ColorPalette:
    __slots__ = ['pal', '_all_numeric']

    def __init__(self):
        self.pal = OrderedDict()
        self._all_numeric = True

    def __repr__(self):
        return str(self.pal)

    def set_num_to_percent(self, ndv=True):
        new_pal = ColorPalette()
        for num, val in self.pal.items():
            if not isinstance(num, str):
                if num < 0:
                    num = 0
                elif num > 100:
                    num = 100
                num = str(num)+'%'
            new_pal.pal[num] = val
        new_pal._all_numeric = False
        if ndv:
            new_pal.pal['nv'] = 0
        return new_pal

    def to_serial_values(self, first=0):
        keys = list(self.pal.keys())
        i = first
        for key in keys:
            if not isinstance(key, str):
                self.pal[i] = self.pal.pop(key)
                i += 1

    def has_percents(self):
        if self._all_numeric:
            False
        for num in self.pal.keys():
            if not isinstance(num, str):
                continue
            is_percent = num.endswith('%')
            if is_percent:
                return True
        return False

    def apply_percent(self, min_val, max_val):
        if self._all_numeric:
            # nothing to do
            return
        all_numeric = True
        new_pal = self.pal.copy()
        for num in self.pal.keys():
            if not isinstance(num, str):
                continue
            is_percent = num.endswith('%')
            if is_percent:
                new_num = num.rstrip('%')
                try:
                    new_num = to_number(new_num)
                    if is_percent:
                        new_num = (max_val - min_val) * new_num * 0.01 + min_val
                    new_pal[new_num] = new_pal.pop(num)
                except ValueError:
                    all_numeric = False
            else:
                all_numeric = False
                continue
        if all_numeric:
            self._all_numeric = True
        self.pal = new_pal

    def read(self, filename_or_strings: ColorPaletteOrPathOrStrings):
        if isinstance(filename_or_strings, ColorPalette):
            self = filename_or_strings.copy()
        else:
            filename, temp_filename = get_file_from_strings(filename_or_strings)
            if Path(filename).suffix.lower() == '.qlr':
                self.read_qlr(filename)
            else:
                self.read_color_file(filename)
            if temp_filename:
                os.remove(temp_filename)

    def read_talos_palette(self, s: str):
        self.pal.clear()
        self._all_numeric = True
        x = s.split(';')
        min_value = to_number(x[0])
        count = int(x[1])
        selected = x[2]
        lock_values = x[3]
        multiplier = to_number(x[4])
        special_draw = x[5]
        interpolate = x[6]
        log_base = to_number(x[8])
        if log_base == 0:
            ln_log_base = None
        else:
            ln_log_base = math.log(log_base)
        j = 8
        for i in range(count):
            name = x[j]
            color = x[j+2]
            color = self.pas_color_to_rgb(color)
            brush = x[j+3]
            key = min_value + i * multiplier
            if ln_log_base:
                key = math.exp(ln_log_base * key)  # == log_base^key
            self.pal[key] = color
            j += 4
        # flags = x[j]
        # pal_version = x[j+1]

    def read_color_file(self, color_filename_or_lines):
        if isinstance(color_filename_or_lines, ColorPalette):
            return self
        elif isinstance(color_filename_or_lines, (Path, str)):
            color_filename_or_lines = open(str(color_filename_or_lines)).readlines()
        elif not isinstance(color_filename_or_lines, Sequence):
            raise Exception('unknwon input {}'.format(color_filename_or_lines))

        self.pal.clear()
        for line in color_filename_or_lines:
            split_line = line.strip().split(' ', maxsplit=1)
            if len(split_line) < 2:
                continue
            try:
                color = self.pal_color_to_rgb(split_line[1])
                key = split_line[0].strip()
            except:
                raise Exception('Error reading palette line: {}'.format(line))
            try:
                key = to_number(key)
            except ValueError:
                # should be percent
                self._all_numeric = False
                pass
            self.pal[key] = color

    def write_color_file(self, color_filename=None):
        if color_filename is None:
            color_filename = tempfile.mktemp(suffix='.txt')
        os.makedirs(os.path.dirname(str(color_filename)), exist_ok=True)
        with open(str(color_filename), mode='w') as fp:
            for key, color in self.pal.items():
                cc = self.color_to_cc(color)
                cc = ' '.join(str(c) for c in cc)
                fp.write('{} {}\n'.format(key, cc))
        return color_filename

    def to_mem_buffer(self):
        s = ''
        for key, color in self.pal.items():
            cc = self.color_to_cc(color)
            cc = ' '.join(str(c) for c in cc)
            s = s + '{} {}\n'.format(key, cc)
        return s

    def read_xml(self, xml_filename, type=None, tag_name=None):
        if tag_name is None:
            if type is None:
                type = Path(xml_filename).suffix
            type = type.lstrip('.').lower()
            if type == 'qlr':
                #             <paletteEntry color="#ffffff" alpha="0" label="0" value="0"/>
                tag_name = "paletteEntry"
            elif type == 'qml':
                #           <item label="-373" color="#d7191c" alpha="255" value="-373"/>
                tag_name = "item"
            else:
                raise Exception('Unknown file type {}'.format(xml_filename))
        self.pal.clear()
        qlr = minidom.parse(str(xml_filename))
        #             <paletteEntry color="#ffffff" alpha="0" label="0" value="0"/>
        color_palette = qlr.getElementsByTagName(tag_name)
        for palette_entry in color_palette:
            color = palette_entry.getAttribute("color")
            if str(color).startswith('#'):
                color = int(color[1:], 16)
            alpha = palette_entry.getAttribute("alpha")
            alpha = int(alpha)
            color = color + (alpha << 8*3)  # * 256**3
            key = palette_entry.getAttribute("value")
            key = to_number(key)
            self.pal[key] = color

    def read_qlr(self, qlr_filename):
        return self.read_xml(qlr_filename, type='qlr')

    def read_qml(self, qml_filename):
        return self.read_xml(qml_filename, type='qml')

    def get_color_table(self, min_val=0, max_val=256, fill_missing_colors=True):
        # create color table
        color_table = gdal.ColorTable()
        if fill_missing_colors:
            keys = list(self.pal.keys())
            vals = list(self.pal.values())
            c = self.color_to_cc(vals[0])
            for key in range(min_val, max_val):
                if key in keys:
                    c = self.color_to_cc(self.pal[key])
                color_table.SetColorEntry(key, c)
        else:
            for key, col in self.pal.items():
                color_table.SetColorEntry(key, self.color_to_cc(col))  # set color for each key
        return color_table

    @staticmethod
    def from_string_list(color_palette_strings):
        res = ColorPalette()
        res.read_color_file(color_palette_strings)
        return res

    @staticmethod
    def format_number(num):
        return num if isinstance(num, str) else '{:.2f}'.format(num)

    @staticmethod
    def format_color(col):
        return col if isinstance(col, str) else '#{:06X}'.format(col)

    @staticmethod
    def color_to_cc(color):
        # if color < 256:
        #     return color
        # else:
            b = byte(color, 0)
            g = byte(color, 1)
            r = byte(color, 2)
            a = byte(color, 3)

            if a < 255:
                return r, g, b, a
            else:
                return r, g, b

    @staticmethod
    def pal_color_to_rgb(cc):
        # r g b a -> argb
        # todo: support color names or just find the gdal implementation of this function...
        # cc = color components
        cc = re.findall(r'\d+', cc)
        try:
            # if not rgb_colors:
            #     return (*(int(c) for c in cc),)
            if len(cc) == 1:
                return int(cc[0])
            elif len(cc) == 3:
                return (((((255 << 8) + int(cc[0])) << 8) + int(cc[1])) << 8) + int(cc[2])
            elif len(cc) == 4:
                return (((((int(cc[3]) << 8) + int(cc[0])) << 8) + int(cc[1])) << 8) + int(cc[2])
            else:
                return 0
        except:
            return 0

    @staticmethod
    def pas_color_to_rgb(col):
        # $CC00FF80
        # $AARRGGBB
        if isinstance(col, str):
            col = str(col).strip('$')
        return int(col, 16)

    @staticmethod
    def from_color_list(color_list):
        res = ColorPalette()
        res.pal.clear()
        res._all_numeric = True
        for key, color in enumerate(color_list):
            res.pal[key] = color
        return res

    @staticmethod
    def from_mcd(mcd_color_list):
        color_list = [int(color.lstrip('#'), 16) for color in mcd_color_list]
        return ColorPalette.from_color_list(color_list)

    @staticmethod
    def get_css4_palette():
        from matplotlib._color_data import CSS4_COLORS as color_dict
        return ColorPalette.from_mcd(color_dict.values())

    @staticmethod
    def get_tableau_palette():
        from matplotlib._color_data import TABLEAU_COLORS as color_dict
        return ColorPalette.from_mcd(color_dict.values())

    @staticmethod
    def get_xkcd_palette():
        from matplotlib._color_data import XKCD_COLORS as color_dict
        return ColorPalette.from_mcd(color_dict.values())


def xml_to_color_file(xml_filename: Path, **kwargs) -> Path:
    xml_filename = Path(xml_filename)
    pal = ColorPalette()
    pal.read_xml(xml_filename, **kwargs)
    color_filename = xml_filename.with_suffix('.txt')
    pal.write_color_file(color_filename)
    return pal, color_filename


def talos_to_color_file(talos_pal: str, color_filename: Path) -> Path:
    pal = ColorPalette()
    pal.read_talos_palette(talos_pal)
    pal = pal.set_num_to_percent(ndv=True)
    pal.write_color_file(color_filename)
    return pal, color_filename


def get_file_from_strings(color_palette: ColorPaletteOrPathOrStrings):
    temp_color_filename = None
    if isinstance(color_palette, ColorPalette):
        temp_color_filename = tempfile.mktemp(suffix='.txt')
        color_filename = temp_color_filename
        color_palette.write_color_file(temp_color_filename)
    elif isinstance(color_palette, (Path, str)):
        color_filename = color_palette
    elif isinstance(color_palette, Sequence):
        temp_color_filename = tempfile.mktemp(suffix='.txt')
        color_filename = temp_color_filename
        with open(temp_color_filename, 'w') as f:
            for item in color_palette:
                f.write(item+'\n')
    else:
        raise Exception('Unknown color palette type {}'.format(color_palette))
    return color_filename, temp_color_filename


def get_color_palette(color_palette_or_path_or_strings: ColorPaletteOrPathOrStrings):
    if color_palette_or_path_or_strings is None:
        return None
    if isinstance(color_palette_or_path_or_strings, ColorPalette):
        pal = color_palette_or_path_or_strings
    else:
        pal = ColorPalette()
        pal.read(color_palette_or_path_or_strings)
    return pal


def get_color_table(color_palette_or_path_or_strings: ColorPaletteOrPathOrStrings):
    pal = get_color_palette(color_palette_or_path_or_strings)
    if pal is None:
        return None
    color_table = pal.get_color_table()
    return color_table


def test_talos_pal(talos_paletts=None):
    if talos_paletts is None:
        talos_paletts = [
            ('percents',
             '0;7;6;0;16.666666666667;0;1;1;0;|;$CC00007F;0;3;2|;$CC0000FF;0;3;2|;$CC00FFFF;0;3;2|;$CC00FF00;0;3;2|;$CCFFFF00;0;3;2|;$CCFF0000;0;3;2|;$CCFF00FF;0;3;2')
        ]
    dir_path = Path(r'sample/color_files')
    for name, talos_pal in talos_paletts:
        path = dir_path / Path(name+'.txt')
        pal, filename = talos_to_color_file(talos_pal, path)
        print(filename, pal)


def test_xml():
    # dir_path = Path('/home/idan/maps/comb')
    dir_path = Path(r'sample/color_files')
    for ext in ['qlr', 'qml']:
        for filename in glob.glob(str(dir_path / '**' / Path('*.'+ext))):
            pal, filename = xml_to_color_file(filename, type=ext)
            print(filename, pal)


if __name__ == "__main__":
    test_talos_pal()
    test_xml()



#http://geoinformaticstutorial.blogspot.it/2012/09/reading-raster-data-with-python-and-gdal.html
#http://www.gis.usu.edu/~chrisg/python/2009/lectures/ospy_slides4.pdf

from osgeo import gdal, osr
from osgeo.gdalconst import *
import numpy as np


def pt2fmt(pt):
    fmttypes = {
        GDT_Byte: 'B',
        GDT_Int16: 'h',
        GDT_UInt16: 'H',
        GDT_Int32: 'i',
        GDT_UInt32: 'I',
        GDT_Float32: 'f',
        GDT_Float64: 'f'
    }
    return fmttypes.get(pt, 'x')


# points_srs == None : pixel/line
# points_srs == False : same as ds
# points_srs == True : EPSG:4326
# points_srs == ... : CloneGeogCS
# points_srs == other srs - transform from given srs
def get_pixel_from_raster(ds, x, y, points_srs=..., ct=None):
    if ds is None:
        raise Exception('Cannot open %s' % raster_filename)

    # Build Spatial Reference object based on coordinate system, fetched from the opened dataset
    if points_srs is None:
        x, y = int(x), int(y)
    else:
        if points_srs is False:
            # input coords are not in same SRS as ds
            if ct is not None:
                srs = osr.SpatialReference()
                srs.ImportFromWkt(ds.GetProjection())
                if points_srs is True:
                    srs = osr.SpatialReference()
                    srs.ImportFromEPSG(4326)  # WGS84 Geo
                elif points_srs is ...:
                    points_srs = srs.CloneGeogCS()  # geographic srs of the same ellipsoid
                if srs.IsSame(points_srs):
                    ct = None
                else:
                    # Transform coordinates from points_srs to raster srs
                    ct = osr.CoordinateTransformation(points_srs, srs)
            if ct is not None:
                (x, y, height) = ct.TransformPoint(x, y)

        # Read geotransform matrix and calculate corresponding pixel coordinates
        geomatrix = ds.GetGeoTransform()
        inv_geometrix = gdal.InvGeoTransform(geomatrix)
        if inv_geometrix is None:
            raise Exception("Failed InvGeoTransform()")

        x, y = int(inv_geometrix[0] + inv_geometrix[1] * x + inv_geometrix[2] * y), \
               int(inv_geometrix[3] + inv_geometrix[4] * x + inv_geometrix[5] * y)

    if x < 0 or x >= ds.RasterXSize or y < 0 or y >= ds.RasterYSize:
        raise Exception('Passed coordinates are not in dataset extent')

    res = ds.ReadAsArray(x, y, 1, 1)

    return res[0][0]


def get_pixel_from_raster_multi(ds, points, points_srs=..., ct=None):
    if len(points) == 0:
        return None
    elif len(points == 1):
        return get_pixel_from_raster(ds, *points[0], points_srs, ct)

    if ds is None:
        raise Exception('Cannot open %s' % raster_filename)

    # Build Spatial Reference object based on coordinate system, fetched from the opened dataset

    if points_srs is False:
        inv_geometrix = None
    else:
        if points_srs is not None:
            # input coords are not in same SRS as ds
            if ct is not None:
                srs = osr.SpatialReference()
                srs.ImportFromWkt(ds.GetProjection())
                if points_srs is ...:
                    points_srs = srs.CloneGeogCS()  # geographic srs of the same ellipsoid
                elif points_srs is True:
                    srs = osr.SpatialReference()
                    srs.ImportFromEPSG(4326)  # WGS84 Geo
                if srs.IsSame(points_srs):
                    ct = None
                else:
                    # Transform coordinates from points_srs to raster srs
                    ct = osr.CoordinateTransformation(points_srs, srs)
            if ct is not None:
                ct.TransformPoints(points)

        # Read geotransform matrix and calculate corresponding pixel coordinates
        geomatrix = ds.GetGeoTransform()
        inv_geometrix = gdal.InvGeoTransform(geomatrix)
        if inv_geometrix is None:
            raise Exception("Failed InvGeoTransform()")

    result = np.empty([len(points)])
    for idx, (x, y) in enumerate(points):
        if inv_geometrix is None:
            x, y = int(x), int(y)
        else:
            x, y = int(inv_geometrix[0] + inv_geometrix[1] * x + inv_geometrix[2] * y), \
                   int(inv_geometrix[3] + inv_geometrix[4] * x + inv_geometrix[5] * y)

        if x < 0 or x >= ds.RasterXSize or y < 0 or y >= ds.RasterYSize:
            raise Exception('Passed coordinates {} are not in dataset extent'.format(idx))

        res = ds.ReadAsArray(x, y, 1, 1)

        result[idx] = res[0][0]

    return result


if __name__ == "__main__":
    raster_filename = r'./data/sample/maps/srtm1_x35_y32.tif'
    lon = (35.1, 35.2, 35.3)
    lat = (32.1, 32.2, 32.3)
    points = np.dstack((lon, lat))[0]
    ds = gdal.Open(raster_filename, gdal.GA_ReadOnly)
    result = get_pixel_from_raster_multi(ds, points)
    print(result)

    result1 = np.empty([len(points)])
    for idx, p in enumerate(points):
        result1[idx] = get_pixel_from_raster(ds, p[0], p[1])
    errs = sum(abs(result1 - result))
    print(errs)



import os
import tempfile
import gdal
from gdalos.gdalos_color import ColorPalette
from gdalos.calc import gdal_calc
from functools import partial
import numpy as np
import copy
from pathlib import Path
from gdalos.calc.discrete_mode import DiscreteMode
from gdalos import gdalos_util
from gdalos.calc import gdal_to_czml


def cont2discrete_array(arr, values, discrete_mode: DiscreteMode, dtype=np.uint8):
    """
    input: continues np array [float | int]
    output: uint8 np array with discrete values
    """
    enumerated_values = list(enumerate(sorted(values)))
    if discrete_mode == DiscreteMode.up:
        # 0 - equal or below first value
        # 1 - above first
        # 2 - above second
        # n - above n
        ret = np.full_like(arr, len(values)-1, dtype=dtype)
        # iterate values from high to low
        for i, val in reversed(enumerated_values[:-1]):
            ret[(arr <= val)] = i
    elif discrete_mode == DiscreteMode.down:
        # 0 - below first value
        # 1 - below second
        # 2 - below third
        # n - above n-1
        ret = np.full_like(arr, 0, dtype=dtype)
        # iterate values from low to high
        for i, val in list(enumerated_values[1:]):
            ret[(arr >= val)] = i
    else:
        raise Exception('unsupported mode {}'.format(discrete_mode))
    return ret


def gdalos_raster_color(filename_or_ds: gdal.Dataset,
                        color_palette: ColorPalette,
                        out_filename: str=None, output_format: str = None,
                        discrete_mode=DiscreteMode.interp) -> gdal.Dataset:
    ds = gdalos_util.open_ds(filename_or_ds)

    if color_palette is None:
        return ds
    if not output_format:
        output_format = 'GTiff' if out_filename else 'MEM'

    if not out_filename:
        out_filename = ''
        if output_format != 'MEM':
            raise Exception('output filename is None')
    if discrete_mode in [DiscreteMode.interp, DiscreteMode.near]:
        temp_color_filename = color_palette.write_color_file()
        dem_options = {
            'options': ['-nearest_color_entry'] if discrete_mode == DiscreteMode.near else [],
            'addAlpha': True,
            'format': output_format,
            'processing': 'color-relief',
            'colorFilename': str(temp_color_filename)}
        ds = gdal.DEMProcessing(str(out_filename), ds, **dem_options)
        os.remove(temp_color_filename)
    elif discrete_mode in [DiscreteMode.up, DiscreteMode.down]:
        color_palette_copy = copy.deepcopy(color_palette)
        if color_palette.has_percents:
            min_max = gdalos_util.get_raster_min_max(ds)
            color_palette_copy.apply_percent(*min_max)

        values = []
        for key in color_palette_copy.pal.keys():
            if not isinstance(key, str):
                values.append(key)
        if not values:
            raise Exception('no absolute values found in the palette')

        f = partial(cont2discrete_array, values=values, discrete_mode=discrete_mode)
        calc_expr = 'f(x)'

        calc_kwargs = dict(x=ds)
        user_namespace = dict(f=f)
        color_palette_copy.to_serial_values()
        color_table = color_palette_copy.get_color_table()

        meta = gdal_to_czml.make_czml_description(color_palette_copy)

        ds = gdal_calc.Calc(
            calc_expr, type=gdal.GDT_Byte, outfile=str(out_filename), format=output_format,
            color_table=color_table, return_ds=True, overwrite=True,
            user_namespace=user_namespace, **calc_kwargs)

        ds.SetMetadataItem(gdal_to_czml.czml_metadata_name, meta)
    if ds is None:
        raise Exception('fail to color')
    return ds


def test_main(path, color_palette):
    for mode in DiscreteMode:
        outfile = path.with_suffix('.{}.tif'.format(mode))
        gdalos_raster_color(path, color_palette, out_filename=outfile, discrete_mode=mode)


def test_dtm():
    path = Path(r'd:\dev\gis\talos_wps\data\sample\maps\srtm1_x35_y32.tif')
    color_file = r'sample\color_files\gradient\rainbow.txt'
    pal = ColorPalette()
    pal.read_color_file(color_file)
    test_main(path, pal)


def test_f1():
    path = Path(r'd:\dev\gis\talos_wps\data\sample\maps\f1.tif')
    talos_pal = r'100;4;0;0;500;0;1;1;0;|;$CC000080;0;3;2|;$CC00FFFF;0;3;2|;$CC00FF00;0;3;2|;$CCFFFF00;0;3;2'
    pal = ColorPalette()
    pal.read_talos_palette(talos_pal)
    test_main(path, pal)


def test_dem_color():
    script_dir = os.path.dirname(os.path.realpath(__file__))
    root_data = os.path.join(script_dir, r'../../data/sample')
    color_palette_filename = os.path.join(root_data, r'color_files/percents.txt')
    raster_filename = os.path.join(root_data, r'maps/srtm1_x35_y32.tif')
    ds = gdalos_util.open_ds(raster_filename)
    out_filename = tempfile.mktemp(suffix='.tif')

    color_palette = ColorPalette()
    color_palette.read_color_file(color_palette_filename)

    ds = gdalos_raster_color(
        ds=ds,
        out_filename=out_filename,
        color_palette=color_palette,
        discrete_mode=DiscreteMode.interp,
        output_format='GTiff')

    print(out_filename)
    print(color_palette)


if __name__ == '__main__':
    # test_dtm()
    test_f1()



import gdal
import numpy as np
from gdalos import gdalos_color, gdalos_trans
from gdalos.gdalos_color import ColorPalette


def test_data_generator(filename, levels=10, of='GTiff', color_palette: ColorPalette=...):
    driver = gdal.GetDriverByName(of)
    gdal_dtype = gdal.GDT_Byte
    np_dtype = np.uint8

    if color_palette is ...:
        color_palette = ColorPalette.get_xkcd_palette()
    color_table = gdalos_color.get_color_table(color_palette)

    for i in range(levels):
        size = 1 << levels-1-i
        shape = (size, size)
        filename_i = filename + '.ovr' * i
        ds = driver.Create(filename_i, xsize=shape[0], ysize=shape[1], bands=1, eType=gdal_dtype)
        pixel_size = 1 << i
        gt = [0, pixel_size, 0, 0, 0, -pixel_size]
        ds.SetGeoTransform(gt)
        bnd = ds.GetRasterBand(1)
        bnd.SetRasterColorTable(color_table)
        bnd.SetRasterColorInterpretation(gdal.GCI_PaletteIndex)
        arr = np.full(shape, i, np_dtype)
        bnd.WriteArray(arr, 0, 0)
        ds.FlushCache()


if __name__ == '__main__':
    filename = r'd:\Maps.temp\test.tif'
    test_data_generator(filename=filename)
    gdalos_trans(filename, overwrite=True)



def gdalos_qt_main():
    from fidget.backend.QtWidgets import QApplication
    from fidget.backend.QtWidgets import QVBoxLayout
    from fidget.core import Fidget
    from fidget.widgets import FidgetMatrix, FidgetMinimal, FidgetQuestion
    from fidget.widgets.__util__ import CountBounds
    from gdalos import gdalos_trans
    from gdalos_qt.gdalos_widget import GdalosWidget

    app = QApplication(
        []
    )  # this app variable is needed, because otherwise python will delete the QApplication
    q = FidgetQuestion(
        FidgetMatrix(
            FidgetMinimal(
                GdalosWidget(make_title=True, make_plaintext=True, make_indicator=True),
                make_title=False,
                make_plaintext=False,
                make_indicator=False,
                initial_value={},
            ),
            rows=CountBounds(1, 1, None),
            columns=1,
            make_plaintext=True,
            make_title=True,
            make_indicator=False,
            layout_cls=QVBoxLayout,
            scrollable=True,
        ),
        flags=Fidget.FLAGS,
    )
    q.show()
    result = q.exec()
    print(result)

    if result.is_ok() and result.value is not None:
        for v in result.value:
            d = dict(v[0])
            d2 = dict()
            for k, v in d.items():
                new_k = k.replace(" ", "_")
                d2[new_k] = d[k]

            print(gdalos_trans(**d2))
    # app.exec_()


if __name__ == "__main__":
    from fidget.backend import prefer

    # the prefer command must be called before importing fidget.backend.QtWidgets
    prefer("PyQt5", True)

    gdalos_qt_main()



import numpy as np
from gdalos.viewshed.viewshed_params import viewshed_thresh, viewshed_ndv, viewshed_comb_ndv, viewshed_comb_multi_val
from gdalos.calc.gdal_calc import AlphaList


def get_by_index(a, index=0):
    """
    this is mainly for testing
    """
    return a[index]


def vs_max(a):
    """
    max value of each pixel
    """
    concatenate = np.stack(a)
    ret = concatenate.max(axis=0)
    return ret


def vs_min(a):
    """
    max value of each pixel
    """
    concatenate = np.stack(a)
    ret = concatenate.min(axis=0)
    return ret


def vs_count(a, threshold=viewshed_thresh):
    """
    count non zero values
    """
    concatenate = np.stack(a)
    a_bools = concatenate > threshold
    ret = a_bools.sum(axis=0, dtype=np.uint8)
    # ret = np.count_nonzero(a_bools, axis=0)  # also works
    return ret


def vs_count_z(a, threshold=viewshed_thresh, in_ndv=viewshed_ndv, out_ndv=viewshed_comb_ndv):
    """
    count non zero values considering ndv values
    """
    concatenate = np.stack(a)
    a_bools = concatenate > threshold
    ret = a_bools.sum(axis=0, dtype=np.uint8)

    # finding all indices that all vales are ndv and set the output to ndv
    a_bools = concatenate != in_ndv
    non_ndv = a_bools.sum(axis=0, dtype=np.uint8)
    ret[non_ndv == 0] = out_ndv

    return ret


def vs_unique(a, threshold=viewshed_thresh, multiple_nz=viewshed_comb_multi_val, all_zero=viewshed_comb_ndv):
    """
    returns indices of rasters with unique values
    """
    nz_count = vs_count(a, threshold)

    ret = np.full_like(a[0], all_zero)
    ret[nz_count > 1] = multiple_nz
    singular = nz_count == 1
    for i, arr in enumerate(a):
        ret[(arr > threshold) & singular] = i
    return ret


def unique_bool(a, multiple_nz=viewshed_comb_multi_val, all_zero=viewshed_comb_ndv):
    a = [np.array(x) for x in a]
    concatenate = np.stack(a)
    nz_count = np.count_nonzero(concatenate, axis=0)
    ret = np.full_like(a[0], all_zero)
    ret[nz_count > 1] = multiple_nz
    singular = nz_count == 1
    for i, arr in enumerate(a):
        ret[(arr != 0) & singular] = i
    return ret


def make_calc_with_operand(filenames, alpha_pattern, operand, **kwargs):
    calc = None
    for filename, alpha in zip(filenames, AlphaList):
        kwargs[alpha] = filename
        alpha1 = alpha_pattern.format(alpha)
        if calc is None:
            calc = alpha1
        else:
            calc = '{}{}{}'.format(calc, operand, alpha1)
    return calc, kwargs


def make_calc_with_func(filenames, alpha_pattern, func_name, **kwargs):
    all_vals = 'a'
    alpha1 = alpha_pattern.format('x')
    calc = '{}({} for x in a)'.format(func_name, alpha1)
    kwargs[all_vals] = filenames
    return calc, kwargs


import math
from numbers import Real

from osgeo import osr, ogr


def get_floats(s):
    # get all floats from a string
    lst = []
    for t in s.split():
        try:
            lst.append(float(t))
        except ValueError:
            pass


def get_float(s):
    # get first float from a string
    res = None
    for t in s.split():
        try:
            res = float(t)
            return res
        except ValueError:
            pass
    return res


def get_number(x):
    if isinstance(x, Real):
        return x
    try:
        val = int(x)
        return val
    except (ValueError, TypeError):
        try:
            val = float(x)
            return val
        except (ValueError, TypeError):
            return None
    # f = float(x)
    # if f.is_integer():
    #     return int(f)
    # else:
    #     return f


def get_zone_from_name(s):
    split_string = s.lower().rsplit("u", 1)
    c = len(split_string)
    if c == 2:
        s = split_string[1]
    elif c != 0:
        return 0
    zone = get_number(s)
    if zone is None:
        return 0
    return zone


def get_zone_center(float_zone):
    zone_center = (float_zone - 30.5) * 6  # == (i-30.5)*6*pi/180
    if zone_center <= -180:
        zone_center += 360
    elif zone_center > 180:
        zone_center -= 360
    return zone_center


def get_zone_by_lon(xdeg: Real, allow_float_zone: bool = False):
    if allow_float_zone:
        return xdeg / 6 + 30.5
    else:
        zone = math.floor(xdeg / 6) + 31
        if zone > 60:
            zone = zone - 60  # Zones 1 - 30
        return zone


def get_datum_and_zone_from_projstring(pjstr):
    srs: osr.SpatialReference = osr.SpatialReference()
    srs.ImportFromProj4(pjstr)
    datum = srs.GetAttrValue('DATUM')
    central_meridian = srs.GetProjParm('central_meridian');
    if central_meridian:
        zone = get_zone_by_lon(central_meridian)
    else:
        zone = None
    return datum, zone


def get_canonic_name(datum, zone):
    if isinstance(datum, str) and datum[0].lower() == "e":
        res = "e50"
    else:
        res = "w84"
    if zone:
        res = res + "u" + str(zone)
    else:
        res = res + "geo"
    return res


def get_utm_zone_extent_points(float_zone, width=10):
    zone_center = get_zone_center(float_zone)
    x_arr = [zone_center - width / 2.0, zone_center + width / 2.0]
    y_arr = [-80, 80]

    extent_points = []
    for x in x_arr:
        for y in y_arr:
            extent_points.append([x, y])
        y_arr.reverse()
    return extent_points


def get_srs_pj_from_ds(ds):
    srs = osr.SpatialReference()
    srs.ImportFromWkt(ds.GetProjection())
    srs_pj4 = srs.ExportToProj4()
    return srs_pj4


def get_srs_pj_from_epsg(epsg=4326):
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    srs_pj4 = srs.ExportToProj4()
    return srs_pj4


def proj_is_equivalent(pj1, pj2):
    if pj1 == pj2:
        return True
    srs1 = osr.SpatialReference()
    srs1.ImportFromProj4(pj1)

    srs2 = osr.SpatialReference()
    srs2.ImportFromProj4(pj2)

    return srs1.IsSame(srs2)


def _srs(srs):
    if isinstance(srs, str):
        srs_ = osr.SpatialReference()
        if srs_.ImportFromProj4(srs) != ogr.OGRERR_NONE:
            raise Exception("ogr error when parsing srs")
        srs = srs_
    return srs


def reproject_coordinates(coords, src_srs, tgt_srs):
    src_srs = _srs(src_srs)
    tgt_srs = _srs(tgt_srs)

    transform = osr.CoordinateTransformation(src_srs, tgt_srs)
    return [transform.TransformPoint(src_x, src_y)[:2] for src_x, src_y in coords]


def get_transform(src_srs, tgt_srs):
    src_srs = _srs(src_srs)
    tgt_srs = _srs(tgt_srs)
    if src_srs.IsSame(tgt_srs):
        return None
    else:
        return osr.CoordinateTransformation(src_srs, tgt_srs)


def get_proj_string(talos_pj, zone=None, **kwargs):
    pj_string, _ = parse_proj_string_and_zone(talos_pj, zone, **kwargs)
    return pj_string


def parse_proj_string_and_zone(talos_pj, zone=None, ED50_towgs84="-87,-98,-121"):
    # r'+proj=tmerc +k=0.9996 +lon_0={} +x_0=500000 +datum=WGS84 +units=m +no_defs'
    # r'+proj=tmerc +k=0.9996 +lon_0={} +x_0=500000 +ellps=intl +towgs84=x,y,z +units=m +no_defs'
    #
    # r'+proj=utm +zone={} +datum=WGS84 +units=m +no_defs'
    # r'+proj=utm +zone={} +ellps=intl +towgs84=x,y,z +units=m +no_defs'
    #
    # r'+proj=latlong +datum=WGS84 +no_defs'
    # r'+proj=latlong +ellps=intl +towgs84=x,y,z +no_defs'

    pj_string = None
    if zone is None:
        number = get_number(talos_pj)
        if number is not None:
            if isinstance(number, int) and number > 100:
                pj_string = '+init=epsg:{}'.format(number)
            else:
                zone = number
        else:
            zone = get_zone_from_name(talos_pj)

    if not pj_string:
        if isinstance(talos_pj, str):
            if talos_pj.startswith("+"):
                pj_string = talos_pj
            elif talos_pj.lower().startswith("epsg"):
                pj_string = '+init=' + talos_pj

    if not pj_string:
        isGeo = zone is None or (zone <= 0)
        if isGeo:
            pj_string = r"+proj=latlong"
        elif float(zone).is_integer():
            pj_string = r"+proj=utm +zone={}".format(zone)
        else:
            pj_string = r"+proj=tmerc +k=0.9996 +lon_0={} +x_0=500000".format(get_zone_center(zone))
        if isinstance(talos_pj, str) and talos_pj[0].lower() == "e":
            pj_string = pj_string + " +ellps=intl +towgs84=" + ED50_towgs84
        else:
            pj_string = pj_string + " +datum=WGS84"
        if not isGeo:
            pj_string = pj_string + " +units=m"
        pj_string = pj_string + " +no_defs"

    return pj_string, zone




import numpy as np
import gdal

gdal_dt_to_np_dt = {gdal.GDT_Byte: np.uint8,
                    gdal.GDT_UInt16: np.uint16,
                    gdal.GDT_Int16: np.int16,
                    gdal.GDT_UInt32: np.uint32,
                    gdal.GDT_Int32: np.int32,
                    gdal.GDT_Float32: np.float32,
                    gdal.GDT_Float64: np.float64}



import tempfile
import math

from osgeo import ogr, osr
import gdal

from gdalos.rectangle import GeoRectangle, get_points_extent
from gdalos import gdalos_util

# Geotransform:
# 0 = x-coordinate of the upper-left corner of the upper-left pixel
# 1 = pixel width,
# 2 = row rotation (typically zero)
# 3 = y-coordinate of the of the upper-left corner of the upper-left pixel
# 4 = column rotation (typically zero)
# 5 = pixel height (typically negative)
# Xgeo = gt(0) + Xpixel * gt(1) + Yline * gt(2)
# Ygeo = gt(3) + Xpixel * gt(4) + Yline * gt(5)

# In case of north up images,
# (GT(2), GT(4)) coefficients are zero,
# (GT(1), GT(5)) is pixel size
# (GT(0), GT(3)) position is the top left corner of the top left pixel of the raster.
# Note that the pixel/line coordinates in the above are from (0.0,0.0) at the top left corner of the top left pixel
# to (width_in_pixels,height_in_pixels) at the bottom right corner of the bottom right pixel.
# The pixel/line location of the center of the top left pixel would therefore be (0.5,0.5).


def calc_dx_dy(extent: GeoRectangle, sample_count: int):
    (min_x, max_x, min_y, max_y) = extent.min_max
    w = max_x - min_x
    h = max_y - min_y
    pix_area = w * h / sample_count
    if pix_area <= 0 or w <= 0 or h <= 0:
        return 0, 0
    pix_len = math.sqrt(pix_area)
    return pix_len, pix_len


def translate_extent(extent: GeoRectangle, transform, sample_count=1000):
    if transform is None:
        return extent
    maxf = float("inf")
    (out_min_x, out_max_x, out_min_y, out_max_y) = (maxf, -maxf, maxf, -maxf)

    dx, dy = calc_dx_dy(extent, sample_count)
    if dx == 0:
        return GeoRectangle.empty()

    y = float(extent.min_y)
    while y <= extent.max_y + dy:
        x = float(extent.min_x)
        while x <= extent.max_x + dx:
            tx, ty, tz = transform.TransformPoint(x, y)
            x += dx
            if not math.isfinite(tz):
                continue
            out_min_x = min(out_min_x, tx)
            out_max_x = max(out_max_x, tx)
            out_min_y = min(out_min_y, ty)
            out_max_y = max(out_max_y, ty)
        y += dy

    return GeoRectangle.from_min_max(out_min_x, out_max_x, out_min_y, out_max_y)


def get_geotransform_and_size(ds):
    return ds.GetGeoTransform(), (ds.RasterXSize, ds.RasterYSize)


def get_points_extent_from_ds(ds):
    geo_transform, size = get_geotransform_and_size(ds)
    points_extent = get_points_extent(geo_transform, *size)
    return points_extent, geo_transform


def dist(p1x, p1y, p2x, p2y):
    return math.sqrt((p2y - p1y) ** 2 + (p2x - p1x) ** 2)


def transform_resolution_p(transform, dx, dy, px, py):
    p1x, p1y, *_ = transform.TransformPoint(px, py + dx, 0)
    p2x, p2y, *_ = transform.TransformPoint(px, py + dy, 0)
    return dist(p1x, p1y, p2x, p2y)


def transform_resolution_old(transform, input_res, extent: GeoRectangle):
    (xmin, xmax, ymin, ymax) = extent.min_max
    [dx, dy] = input_res
    out_res_x = min(
        transform_resolution_p(transform, 0, dy, xmin, ymin),
        transform_resolution_p(transform, 0, -dy, xmin, ymax),
        transform_resolution_p(transform, 0, -dy, xmax, ymax),
        transform_resolution_p(transform, 0, dy, xmax, ymin),
    )
    if (ymin > 0) and (ymax < 0):
        out_res_x = min(
            out_res_x,
            transform_resolution_p(transform, 0, dy, xmin, 0),
            transform_resolution_p(transform, 0, dy, xmax, 0),
        )
    out_res_x = round_to_sig(out_res_x, -1)
    out_res = (out_res_x, -out_res_x)
    return out_res


def transform_resolution(
    transform, input_res, extent: GeoRectangle, equal_res=..., sample_count=1000):
    dx, dy = calc_dx_dy(extent, sample_count)

    calc_only_res_y = equal_res is ...
    out_x = []
    out_y = []
    y = float(extent.min_y)
    while y <= extent.max_y:
        x = float(extent.min_x)
        while x < extent.max_x:
            out_y.append(transform_resolution_p(transform, 0, input_res[1], x, y))
            if not calc_only_res_y:
                out_x.append(transform_resolution_p(transform, input_res[0], 0, x, y))
            x += dx
        y += dy

    if calc_only_res_y:
        out_y.sort()
        out_x = out_y
    if equal_res:
        out_y.extend(out_x)
        out_y.sort()
        out_x = out_y
    else:
        out_x.sort()
        out_y.sort()
    out_r = [out_x, out_y]

    # choose the median resolution
    out_res = [round_to_sig(r[round(len(r) / 2)], -1) for r in out_r]
    out_res[1] = -out_res[1]
    return out_res


# todo we need to get rid of this
def round_to_sig(d, extra_digits=-5):
    if (d == 0) or math.isnan(d) or math.isinf(d):
        return 0
    if abs(d) > 1e-20:
        digits = int(math.floor(math.log10(abs(d) + 1e-20)))
    else:
        digits = int(math.floor(math.log10(abs(d))))
    digits = digits + extra_digits
    return round(d, -digits)


def get_vec_extent(lyr):
    result = None
    for feature in lyr:
        geom = feature.GetGeometryRef()
        envelope = geom.GetEnvelope()
        r = GeoRectangle.from_min_max(*envelope)
        if result is None:
            result = r
        else:
            result = result.union(r)
    return result


def get_extent(filename_or_ds) -> GeoRectangle:
    ds = gdalos_util.open_ds(filename_or_ds)
    gt, size = get_geotransform_and_size(ds)
    return GeoRectangle.from_geotransform_and_size(gt, size)


def calc_geo_offsets(src_gt, src_size, dst_gt, dst_size):
    src = GeoRectangle.from_geotransform_and_size_to_pix(src_gt, src_size)
    dst = GeoRectangle.from_geotransform_and_size_to_pix(dst_gt, dst_size)
    offset = (dst.x - src.x, dst.y - src.y)
    src_offset = (max(0, offset[0]), max(0, offset[1]))
    dst_offset = (max(0, -offset[0]), max(0, -offset[1]))
    return src_offset, dst_offset


def calc_geotransform_and_dimensions(geotransforms, dimensions, input_extent: GeoRectangle = None):
    # extents differ, but pixel size and rotation are the same.
    # we'll make a union or an intersection
    if geotransforms is None or len(geotransforms) != len(dimensions):
        raise Exception('Error! GeoTransforms and Dimensions have different lengths!')
    if isinstance(input_extent, GeoRectangle):
        gt = geotransforms[0]
        out_extent = input_extent.align(gt)
    else:
        out_extent: GeoRectangle = None
        is_union = input_extent == 2
        for gt, size in zip(geotransforms, dimensions):
            extent = GeoRectangle.from_geotransform_and_size(gt, size)
            out_extent = extent if out_extent is None else \
                out_extent.union(extent) if is_union else out_extent.intersect(extent)

    if out_extent is None or out_extent.is_empty():
        return None, None, None
    else:
        pixel_size = (gt[1], gt[5])
        pix_extent = out_extent.to_pixels(pixel_size)
        gt = (out_extent.left,
              gt[1], gt[2],
              out_extent.up,
              gt[4], gt[5])
    return gt, (math.ceil(pix_extent.w), math.ceil(pix_extent.h)), out_extent


def make_temp_vrt(ds, extent: GeoRectangle):
    # ds = gdalos_util.open_ds(filename)
    # extent = gdalos_extent.get_extent(ds)
    options = gdal.BuildVRTOptions(outputBounds=(extent.min_x, extent.min_y, extent.max_x, extent.max_y))
    # hideNodata=True, separate=False)
    # vrt_filename = filename + ".vrt"
    vrt_filename = tempfile.mktemp(suffix='.vrt')
    vrt_ds = gdal.BuildVRT(vrt_filename, ds, options=options)
    if vrt_ds is None:
        raise Exception("Error! cannot create vrt. Cannot proceed")
    # close_and_reopen = False
    # if close_and_reopen:
    #     vrt_ds = None
    #     vrt_ds = gdalos_util.open_ds(vrt_filename)
    return vrt_filename, vrt_ds


def make_temp_vrt_old(filename, ds, data_type, projection, bands_count, gt, dimensions, ref_gt, ref_dimensions):
    drv = gdal.GetDriverByName("VRT")
    dt = gdal.GetDataTypeByName(data_type)

    # vrt_filename = filename + ".vrt"
    vrt_filename = tempfile.mktemp(suffix='.vrt')
    vrt_ds = drv.Create(vrt_filename, ref_dimensions[0], ref_dimensions[1], bands_count, dt)
    vrt_ds.SetGeoTransform(ref_gt)
    vrt_ds.SetProjection(projection)

    src_offset, dst_offset = calc_geo_offsets(gt, dimensions, ref_gt, ref_dimensions)
    if src_offset is None:
        raise Exception("Error! The requested extent is empty. Cannot proceed")

    source_size = dimensions

    for j in range(1, bands_count + 1):
        band = vrt_ds.GetRasterBand(j)
        inBand = ds.GetRasterBand(j)
        myBlockSize = inBand.GetBlockSize()

        myOutNDV = inBand.GetNoDataValue()
        if myOutNDV is not None:
            band.SetNoDataValue(myOutNDV)
        ndv_out = '' if myOutNDV is None else '<NODATA>%i</NODATA>' % myOutNDV

        source_xml = '<SourceFilename relativeToVRT="1">%s</SourceFilename>' % filename + \
                     '<SourceBand>%i</SourceBand>' % j + \
                     '<SourceProperties RasterXSize="%i" RasterYSize="%i" DataType=%s BlockXSize="%i" BlockYSize="%i"/>' % \
                     (*source_size, dt, *myBlockSize) + \
                     '<SrcRect xOff="%i" yOff="%i" xSize="%i" ySize="%i"/>' % \
                     (*src_offset, *source_size) + \
                     '<DstRect xOff="%i" yOff="%i" xSize="%i" ySize="%i"/>' % \
                     (*dst_offset, *source_size) + \
                     ndv_out
        source = '<ComplexSource>' + source_xml + '</ComplexSource>'
        band.SetMetadataItem("source_%i" % j, source, 'new_vrt_sources')
        band = None  # close band
    return vrt_filename, vrt_ds

# def GetExtent(gt,cols,rows):
#     ''' Return list of corner coordinates from a gt
#
#         @type gt:   C{tuple/list}
#         @param gt: gt
#         @type cols:   C{int}
#         @param cols: number of columns in the dataset
#         @type rows:   C{int}
#         @param rows: number of rows in the dataset
#         @rtype:    C{[float,...,float]}
#         @return:   coordinates of each corner
#     '''
#     ext=[]
#     xarr=[0, cols]
#     yarr=[0, rows]
#
#     for px in xarr:
#         for py in yarr:
#             x=gt[0]+(px*gt[1])+(py*gt[2])
#             y=gt[3]+(px*gt[4])+(py*gt[5])
#             ext.append([x, y])
#         yarr.reverse()
#     return ext




from gdalos import gdalos_trans
from gdalos import GeoRectangle
from pathlib import Path
from gdalos.viewshed.viewshed_grid_params import ViewshedGridParams


def calc_extent(oxy, grid_range, interval, md, frame):
    # i is for x-y;
    # j is for min and max values of the grid
    # k is for the sign (to subtract or to add) the max_range
    minmax = [oxy[i] + grid_range[j] * interval + k * (md + frame) for i in (0, 1) for j, k in zip((0, -1), (-1, 1))]
    full_extent = GeoRectangle.from_min_max(*minmax)
    return full_extent


def viewshed_run(vp: ViewshedGridParams, output_path, input_filename):
    import gdal
    from gdalos import gdalos_util

    ds = gdalos_util.open_ds(input_filename)
    band: gdal.Band = ds.GetRasterBand(1)

    if band is None:
        raise Exception('band number out of range')

    arr = vp.get_array()
    for vp1 in arr:
        filename = output_path / (vp1.name + '.tif')

        vals = vp1.get_as_gdal_params()
        dest = gdal.ViewshedGenerate(band, 'GTiff', str(filename), None, **vals)

        if dest is None:
            raise Exception('error occurred')

        del dest


if __name__ == "__main__":
    input_filename = Path(r'd:\Maps\w84u36\dtm\SRTM1_hgt.x[27.97,37.98]_y[27.43,37.59].cog.tif')
    output_path = Path(r'd:\dev\gis\maps')

    # dir_path = Path('/home/idan/maps')
    # input_filename = dir_path / Path('srtm1_x35_y32.tif')
    # output_path = dir_path / Path('comb')

    srtm_filename = Path(output_path) / Path('srtm1_36_sample.tif')

    vp = ViewshedGridParams()

    make_map = True
    if make_map:
        frame = 500
        full_extent = calc_extent(vp.oxy, vp.grid_range, vp.interval, vp.max_r, frame)
        gdalos_trans(input_filename, extent=full_extent, warp_srs=36, extent_in_4326=False, out_filename=srtm_filename)

    viewshed_run(vp, output_path, srtm_filename)



import math
import numpy as np
from functools import partial
import gdal
from pathlib import Path
from gdalos.calc import gdal_calc, gdal_numpy
from gdalos import gdalos_util, gdalos_trans, GeoRectangle
from numbers import Real


def autoscale(bnd, np_dtype, possible_scale_values=(0.1, 0.15, 0.2, 0.25, 0.3)):
    bnd.ComputeStatistics(0)
    max_band_val = bnd.GetMaximum()
    max_dt_value = np.iinfo(np_dtype).max
    scale = max_band_val / max_dt_value
    if possible_scale_values is None:
        scale = math.ceil(scale*100)/100
    else:
        for v in possible_scale_values:
            if scale <= v:
                scale = v
                break
    return scale


def scale_np_array(arr, factor: Real, in_ndv, out_ndv, dtype):
    """
        input: arr: numpy array, factor
        returns a numpy array = arr*factor with a given dtype.
    """

    # ret = np.full_like(arr, out_ndv, dtype=dtype)
    # ret[arr != in_ndv] = dtype(arr * scale)

    ret = np.full_like(arr, out_ndv, dtype=dtype)
    np.multiply(arr, factor, out=ret, casting='unsafe')
    ret[arr == in_ndv] = out_ndv

    return ret


def scale_raster(filename_or_ds, d_path, gdal_dt=gdal.GDT_Int16,
                 hide_nodata = True, in_ndv=..., out_ndv=..., scale=0, **kwargs):
    ds = gdalos_util.open_ds(filename_or_ds)
    if in_ndv is ...:
        in_ndv = gdalos_util.get_nodatavalue(ds)
    if out_ndv is ...:
        out_ndv = gdal_calc.DefaultNDVLookup[gdal_dt]
    np_dtype = gdal_numpy.gdal_dt_to_np_dt[gdal_dt]
    if not scale:
        scale = autoscale(ds.GetRasterBand(1), np_dtype)
    f = partial(scale_np_array, factor=1 / scale, in_ndv=in_ndv, out_ndv=out_ndv, dtype=np_dtype)
    calc_expr = 'f(x)'
    calc_kwargs = dict(x=ds)
    user_namespace = dict(f=f)
    creation_options = gdalos_util.get_creation_options()

    ds = gdal_calc.Calc(
        calc_expr, outfile=str(d_path), hideNodata=hide_nodata, NoDataValue=out_ndv,
        overwrite=True, return_ds=True,
        user_namespace=user_namespace, creation_options=creation_options, **kwargs, **calc_kwargs)

    for i in range(ds.RasterCount):
        ds.GetRasterBand(i + 1).SetScale(scale)

    return ds


def make_test_input():
    input_file = r'd:\Maps\w84u36\dtm\SRTM1_hgt.x[27.97,37.98]_y[27.43,37.59].cog.tif'
    extent = GeoRectangle.from_min_max(670000, 700000, 3600000, 3630000)
    gdalos_trans(input_file, extent=extent, extent_in_4326=False)


def make_test_input2():
    input_file = Path(r'd:\Maps\w84geo\dtm\SRTM1_hgt.tif.new.cog.tif')
    # input_file = r'd:\Maps\w84u36\dtm\SRTM1_hgt.x[27.97,37.98]_y[27.43,37.59].cog.tif'
    extent = GeoRectangle.from_min_max(34, 35, 32, 33)
    gdalos_trans(input_file, warp_srs=36, extent=extent, extent_in_4326=True, out_path=r'd:\Maps.temp', out_path_with_src_folders=False)


def test_scale_raster(input_file):
    scale = 0
    out_dst = Path(input_file).with_suffix('.scale_{}.tif'.format(scale))
    scale_raster(input_file, out_dst, scale=scale)
    gdalos_trans(out_dst)
    return out_dst


if __name__ == '__main__':
    # make_test_input2()
    input_file = Path(r'd:\Maps\w84u36\dtm\SRTM1_hgt.x[27.97,37.98]_y[27.43,37.59].cog.tif.x[34.8,35.16]_y[32.51,32.81].cog.tif')
    # input_file = Path(r'd:\Maps.temp\SRTM1_hgt.tif.new.cog.tif.w84u36.[31.0, -31.0].x[33.99,35.06]_y[31.99,33.05].cog.tif')
    # input_file = Path(r'd:\Maps\w84u36\dtm\SRTM1_hgt.x[27.97,37.98]_y[27.43,37.59].cog.tif')
    # input_file = r"d:\Maps.raw\osm\ready\w84u36\SRTM1_hgt_ndv0.cog.tif.new.cog.tif.w84u36.[31.0, -31.0].x[26.85,39.55]_y[19.92,40.4].cog.tif"
    test_scale_raster(input_file)



import math
from collections import defaultdict
from typing import List

import gdal
import re
import os
from pathlib import Path
import shutil

from gdalos.rectangle import GeoRectangle
from gdalos.gdalos_extent import get_extent
from gdalos.gdalos_util import open_ds, OpenDS


class AutoRepr(object):
    def __repr__(self):
        items = ("%s = %r" % (k, v) for k, v in self.__dict__.items())
        return "<%s: {%s}>" % (self.__class__.__name__, ', '.join(items))


class RasterOverview(AutoRepr):
    def __init__(self, path: Path, ovr_idx=0):
        self.path = path
        if ovr_idx is None:
            ovr_idx = 0
        self.ovr_idx = ovr_idx
        ds = self.get_ds()
        # self.bnd = self.ds.GetRasterBand(1)
        # self.ovr_count = self.bnd.GetOverviewCount()
        self.extent = get_extent(ds)
        self.geo_transform = ds.GetGeoTransform()
        self.srs = ds.GetSpatialRef()
        self.resx = self.geo_transform[1]
        self.resy = self.geo_transform[5]
        del ds

    def get_level(self, min_r):
        level = self.resx / min_r
        level = math.pow(2, round(math.log2(level)))
        return level

    def get_ds(self):
        return open_ds(self.path, ovr_idx=self.ovr_idx)

    # def __repr__(self):
    #     return str([str(ovr) for ovr in self.o])


class RasterOverviewList(AutoRepr):
    def __init__(self, path: Path):
        self.path = path
        with OpenDS(self.path) as ds:
            bnd = ds.GetRasterBand(1)
            ovr_count = bnd.GetOverviewCount()
        self.o = []
        for ovr in range(0, ovr_count+1):
            self.o.append(RasterOverview(self.path, ovr))

    def __repr__(self):
        return '\n'.join([str(ovr) for ovr in self.o])


def print_ros(ros: List[RasterOverview]):
    x = '\n'.join([str(ovr) for ovr in ros])
    print(x)


def make_vrt_with_multiple_extent_overviews_from_raster_overview_list(ros: List[RasterOverview], vrt_filename, **kwargs):
    if not ros:
        return None
    print_ros(ros)

    extent = ros[0].extent
    # r = set()
    min_r = math.inf
    max_r = -math.inf
    srs = None
    path = None
    for ro in ros:
        extent = ro.extent.union(extent)
        x = ro.resx
        # r.add(x)
        min_r = min(min_r, x)
        max_r = max(max_r, x)
        if srs is None:
            srs = ro.srs
            path = ro.path
        elif not srs.IsSame(ro.srs):
            raise Exception('srs not the same, please separate by srs:\n{}\n{}'.format(path, ro.path))

    fact = max_r / min_r

    print(extent)
    print('res ({}) - ({})'.format(min_r, max_r))

    # levels = set()
    # for ro in ros:
    #     levels.add(ro.get_level(min_r))
    # levels = sorted(levels)
    # print('levels: {}'.format(levels))

    d = defaultdict(list)
    for ro in ros:
        level = ro.get_level(min_r)
        d[level].append(ro)

    for key in sorted(d.keys()):
        ros = d[key]
        ros = sorted(ros, key=lambda ro: ro.extent.area, reverse=True)
        d[key] = ros

    best_ros = []
    for key in sorted(d.keys()):
        ros = d[key]
        best_ros.append(ros[0])
        for ro in ros:
            print('{}: {}'.format(key, ro))
    ros = best_ros

    print('best ros:')
    print_ros(ros)
    for idx, ro in enumerate(ros):
        # if ro.ovr < 0:
        #     continue
        print('{}: {}'.format(idx, ro.path))
        single_src_vrt = ro.path.with_suffix('.{}.vrt'.format(ro.ovr_idx))
        # single_src_vrt = (ro.path.with_name('vrt') / ro.path.name).with_suffix('.{}.vrt'.format(ro.ovr))
        make_ros_vrt([ro], extent, single_src_vrt)
        ro.path = single_src_vrt
        ro.ovr_idx = 0
    return make_ros_vrt_overviews(ros, extent, vrt_filename, **kwargs)


def vrt_add_overviews(ros: List[RasterOverview], filename_in: Path, filename_out: Path):
    # <SourceFilename relativeToVRT="1">0.vrt</SourceFilename>

    rsrc_bnd = re.compile('(\s*)<SourceBand>\d+</SourceBand>\s*')
    rSource = re.compile('(\s*)</(Simple|Complex)Source>\s*')

    # <Overview>
    #   <SourceFilename relativeToVRT="1">2\2.vrt</SourceFilename>
    #   <SourceBand>1</SourceBand>
    # </Overview>
    dir_name = Path(os.path.dirname(filename_out))
    with open(filename_in) as f, open(filename_out, "w") as f2:
        for line in f:
            f2.write(line)
            m = rsrc_bnd.match(line)
            if m:
                src_bnd = '\t' + line
            else:
                m = rSource.match(line)
                if m:
                    for ro in ros:
                        ovr_filename = ro.path
                        ovr_path = (dir_name / Path(ovr_filename)).resolve()
                        ovr_filename = os.path.relpath(ovr_path, dir_name)
                        if ovr_path.is_file():
                            f2.write('\t<Overview>\n')
                            f2.write('\t\t<SourceFilename relativeToVRT="1">{}</SourceFilename>\n'.format(str(ovr_filename)))
                            f2.write(src_bnd)
                            f2.write('\t</Overview>\n')


def vrt_fix_openoptions(ros: List[RasterOverview], filename_in: Path, filename_out: Path):
    # for some reason open options are not written to the vrt.
    # I'll open a bug report, in the mean time I wrote this workaround
    p = re.compile('<SourceFilename relativeToVRT="(.*)">(.*)</SourceFilename>')
    dir_name = Path(os.path.dirname(filename_out))
    with open(filename_in) as f, open(filename_out, "w") as f2:
        for line in f:
            f2.write(line)
            m = p.search(line)
            if m:
                relative = m.group(1)
                source_filename = m.group(2)
                #   <OpenOptions>
                #       <OOI key="OVERVIEW_LEVEL">6</OOI>
                #   </OpenOptions>
                ovr_idx = ...
                for ro in ros:
                    if relative:
                        source_filename = (dir_name / source_filename).resolve()
                    if ro.path == source_filename:
                        ovr_idx = ro.ovr_idx
                        if ovr_idx > 0:
                            f2.write(
                                '\t<OpenOptions>\n\t\t<OOI key="OVERVIEW_LEVEL">{}</OOI>\n\t</OpenOptions>\n'.format(
                                    ovr_idx-1))
                        break
                if ovr_idx is ...:
                    raise Exception('SourceFilename: {} not found'.format(source_filename))


def make_ros_vrt_overviews(ros: List[RasterOverview], extent: GeoRectangle, vrt_filename: Path, add_overviews=True, return_ds=False):
    options = gdal.BuildVRTOptions(outputBounds=(extent.min_x, extent.min_y, extent.max_x, extent.max_y))
    vrt_ds = gdal.BuildVRT(str(vrt_filename), ros[0].get_ds(), options=options)
    if vrt_ds is None:
        raise Exception("Error! cannot create vrt. Cannot proceed")
    if add_overviews:
        vrt_ds = None
        vrt_filename_temp = vrt_filename.with_suffix('.tmp.vrt')
        shutil.move(vrt_filename, vrt_filename_temp)
        vrt_add_overviews(ros[1:], filename_in=vrt_filename_temp, filename_out=vrt_filename)
        os.remove(vrt_filename_temp)
        if return_ds:
            vrt_ds = open_ds(vrt_filename)
    if return_ds:
        return vrt_ds
    else:
        return vrt_filename


def make_ros_vrt(ros: List[RasterOverview], extent: GeoRectangle, vrt_filename: Path, fix_open_options=True):
    options = gdal.BuildVRTOptions(outputBounds=(extent.min_x, extent.min_y, extent.max_x, extent.max_y))
    # dir_name = Path(os.path.dirname(str(vrt_filename)))
    # ds_list = [os.path.relpath(ro.path, dir_name) for ro in ros]
    ds_list = [ro.get_ds() for ro in ros]
    os.makedirs(os.path.dirname(str(vrt_filename)), exist_ok=True)
    vrt_ds = gdal.BuildVRT(str(vrt_filename), ds_list, options=options)
    if vrt_ds is None:
        raise Exception("Error! cannot create vrt. Cannot proceed")
    if fix_open_options:
        vrt_ds = None
        vrt_filename_temp = vrt_filename.with_suffix('.tmp.vrt')
        shutil.move(vrt_filename, vrt_filename_temp)
        vrt_fix_openoptions(ros, filename_in=vrt_filename_temp, filename_out=vrt_filename)
        os.remove(vrt_filename_temp)
        vrt_ds = open_ds(vrt_filename)
    return vrt_ds


def make_overviews_vrt(paths: List[Path], vrt_filename=None, **kwargs):
    if not paths:
        return None
        # raise Exception ('no files are given')
    paths = [Path(path) for path in paths]
    print(paths)
    if vrt_filename is None:
        first = paths[0]
        vrt_filename = first.with_suffix('.super.vrt')
    o = []
    for path in paths:
        r = RasterOverviewList(path)
        o.extend(r.o)
    return make_vrt_with_multiple_extent_overviews_from_raster_overview_list(o, vrt_filename=vrt_filename, **kwargs)


def make_overviews_vrt_dir(path: Path, pattern='*.tif', outside=True, inside=True, **kwargs):
    path = Path(path)
    paths = list(path.glob(pattern=pattern))

    result = None
    vrt_filenames=[]
    if outside:
        vrt_filenames.append(path.with_suffix('.vrt'))
    if inside:
        dir_name = Path(os.path.basename(str(path)))
        vrt_filenames.append(path / dir_name.with_suffix('.vrt'))
    for vrt_filename in vrt_filenames:
        result = make_overviews_vrt(paths=paths, vrt_filename=vrt_filename, **kwargs)
    return result


def make_overviews_vrt_super_dir(super_path: Path, **kwargs):
    for path in Path(super_path).glob(pattern='*'):
        if path.is_dir():
            make_overviews_vrt_dir(path, **kwargs)


if __name__ == '__main__':
    # make_overviews_vrt_super_dir(r'd:\Maps.raw\osm')
    # make_overviews_vrt_super_dir(r'd:\Maps\temp\x')
    # make_overviews_vrt_super_dir(r'd:\Maps\w84geo\topo')
    # make_overviews_vrt_super_dir(r'd:\Maps.raw\osm')
    print('done!')



import datetime
import logging
import os
import time
import tempfile
from enum import Enum, auto
from numbers import Real
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, TypeVar, Union

from osgeo import gdal
from gdalos import gdalos_util, gdalos_logger, gdalos_extent, projdef
from gdalos.__util__ import with_param_dict
from gdalos.rectangle import GeoRectangle, make_partitions
from gdalos_data.__data__ import __version__
from gdalos.gdalos_types import GdalOutputFormat, OvrType, enum_to_str, GdalResamplingAlg


def print_time_now(logger):
    logger.info("Current time: {}".format(datetime.datetime.now()))


class RasterKind(Enum):
    photo = auto()
    pal = auto()
    dtm = auto()

    @classmethod
    def guess(cls, band_types_or_filename_or_ds):
        if isinstance(band_types_or_filename_or_ds, list):
            band_types = band_types_or_filename_or_ds
        else:
            band_types = gdalos_util.get_band_types(band_types_or_filename_or_ds)
        if len(band_types) == 0:
            raise Exception("no bands in raster")

        if band_types[0] == "Byte":
            if len(band_types) in (3, 4):
                return cls.photo
            elif len(band_types) == 1:
                return cls.pal
            else:
                raise Exception("invalid raster band count")
        elif len(band_types) == 1:
            return cls.dtm

        raise Exception("could not guess raster kind")


def resampling_alg_by_kind(kind, expand_rgb=False, fast_mode=False) -> GdalResamplingAlg:
    if kind == RasterKind.pal and not expand_rgb:
        if fast_mode:
            return GdalResamplingAlg.nearest
        else:
            return GdalResamplingAlg.mode
    else:
        if fast_mode:
            return GdalResamplingAlg.average
        else:
            return GdalResamplingAlg.cubic


def do_skip_if_exists(out_filename, overwrite, logger=None):
    verbose = logger is not None and logger is not ...
    skip = False
    if os.path.isfile(out_filename):
        if not overwrite:
            skip = True
            if verbose:
                logger.warning('file "{}" exists, skip!'.format(out_filename))
        else:
            if verbose:
                logger.warning('file "{}" exists, deleting...!'.format(out_filename))
            os.remove(out_filename)
    return skip


def print_progress_from_to(r0, r1):
    # print(str(round(r1*100)) + '%', end=" ")
    i0 = 0 if (r0 is None) or (r0 > r1) else round(r0 * 100) + 1
    i1 = round(r1 * 100) + 1
    for i in range(i0, i1):
        print(str(i) if i % 5 == 0 else ".", end="", flush=True)
    if r1 >= 1:
        print("% done!")


def print_progress_callback(print_progress):
    if print_progress:
        if print_progress is ...:
            last = None

            def print_progress(prog, *_):
                nonlocal last

                r0 = last
                r1 = prog
                print_progress_from_to(r0, r1)
                last = prog

    return print_progress


T = TypeVar("T")
MaybeSequence = Union[T, Sequence[T]]
warp_srs_base = Union[str, int, Real]
default_multi_byte_nodata_value = -32768


# overviews numbers explained:
# overview_count=n means that the raster has n+1 rasters inside it. the base raster + n overviews.
# gdal numbers the overviews as 0..n-1, first overview as 0 meaning the base raster has no overview number
# I wanted to give the base raster also a number, which by this logic makes source it would be -1
# I find it a bit confusing to number the rasters as -1..n-1.
# so I define src_ovr=0 to be the base raster and 0..n to be the overviews
# I also define negative overview numbers as follows: ovr_idx<0 == overview_count+ovr_idx+1
# So for raster that has overview_count=3 holds 4 rasters which are numberd 0..3. (-1)->3; (-2)->2, (-3)->1, (-4)->0


@with_param_dict("all_args")
def gdalos_trans(
        filename: MaybeSequence[str],
        filenames_expand: Union[type(...), bool] = ...,
        out_filename: str = None,
        out_path: str = None,
        return_ds: Union[type(...), bool] = ...,
        out_path_with_src_folders: str = True,
        overwrite=False,
        cog: Union[type(...), bool] = ...,
        prefer_2_step_cog: Union[type(...), bool] = ...,
        write_info: bool = True,
        write_spec: bool = True,
        multi_file_as_vrt: bool = False,
        of: MaybeSequence[Union[type(...), GdalOutputFormat, str]] = ...,
        outext: str = ...,
        tiled: bool = True,
        block_size: int = ...,
        big_tiff: str = ...,
        config_options: dict = None,
        open_options: dict = None,
        common_options: dict = None,
        creation_options: dict = None,
        translate_options: dict = None,
        warp_options: dict = None,
        extent: Union[Optional[GeoRectangle], List[GeoRectangle]] = None,
        extent_in_4326: bool = True,
        extent_crop_to_minimal: bool = True,
        src_win=None,
        cutline: Optional[Union[str, List[str]]] = None,
        out_res: Optional[Union[type(...), Real, Tuple[Real, Real]]] = None,  # None: gdalos decides; ...: gdal decides
        warp_srs: MaybeSequence[warp_srs_base] = None,
        warp_scale: Union[Real, Tuple[Real, Real]] = 1,  # https://github.com/OSGeo/gdal/issues/2810
        warp_error_threshold: Optional[Real] = 0,  # [None|0]: [linear approximator|exact] coordinate reprojection
        ovr_type: Optional[OvrType] = OvrType.auto_select,
        ovr_idx: Optional[Union[type(...), int]] = None,  # ovr_idx=0 is the base raster; 1 is the first overview
        keep_src_ovr_suffixes: bool = True,
        dst_ovr_count: Optional[int] = None,
        dst_nodatavalue: Optional[Union[type(...), Real]] = None,  # None -> don't change; ... -> change only for DTM
        src_nodatavalue: Optional[Union[type(...), Real]] = ...,  # None -> use minimum; ... -> use original
        hide_nodatavalue: bool = False,
        kind: RasterKind = None,
        resampling_alg: Union[type(...), None, GdalResamplingAlg, str] = ...,
        lossy: bool = None,
        expand_rgb: bool = False,
        quality: Real = ...,
        keep_alpha: bool = ...,
        sparse_ok: bool = True,
        final_files: list = None,
        ovr_files: list = None,
        aux_files: list = None,
        temp_files: list = None,
        delete_temp_files=True,
        partition=None,
        print_progress=...,
        logger=...,
        *,
        all_args: dict = None,
):
    print(all_args)

    if final_files is None:
        final_files = []
    if ovr_files is None:
        ovr_files = []
    if aux_files is None:
        aux_files = []
    if temp_files is None:
        temp_files = []

    if isinstance(kind, str):
        kind = RasterKind[kind]
    if isinstance(ovr_type, str):
        ovr_type = OvrType[ovr_type]
    elif ovr_type is ...:
        ovr_type = OvrType.auto_select

    if isinstance(partition, int):
        partition = None if partition <= 1 else make_partitions(partition)
        all_args["partition"] = partition

    key_list_arguments = [
        "filename",
        "extent",
        "warp_srs",
        # "cutline"
        "of",
        "expand_rgb",
        "partition",
    ]
    for key in key_list_arguments:
        val = all_args[key]
        do_expand_glob = filenames_expand if (key == "filename") else False
        val = gdalos_util.flatten_and_expand_file_list(val, do_expand_glob=do_expand_glob)
        if key == "filename":
            if gdalos_util.is_list_like(val) and multi_file_as_vrt:
                vrt_path = gdalos_vrt(
                    val,
                    filenames_expand=False,
                    resampling_alg=resampling_alg,
                    kind=kind,
                    overwrite=overwrite,
                    logger=logger,
                )
                if vrt_path is None:
                    raise Exception(
                        "failed to create a vrt file: {}".format(vrt_path)
                    )  # failed?
                temp_files.append(vrt_path)
                val = vrt_path
            filename = val

        if gdalos_util.is_list_like(val):
            # input argument is a list, recurse over its values
            all_args_new = all_args.copy()
            ret_code = None
            for idx, v in enumerate(val):
                print("iterate over {} ({}/{}) - {}".format(key, idx + 1, len(val), v))
                all_args_new[key] = v
                all_args_new["temp_files"] = []
                all_args_new["final_files"] = []
                all_args_new["ovr_files"] = []
                all_args_new["aux_files"] = []
                ret_code = gdalos_trans(**all_args_new)
                temp_files.extend(all_args_new["temp_files"])
                final_files.extend(all_args_new["final_files"])
                ovr_files.extend(all_args_new["ovr_files"])
                aux_files.extend(all_args_new["aux_files"])
                # if ret_code is None:
                #     break  # failed?
            return ret_code
        else:
            all_args[key] = val  # adding the default parameters

    if not filename:
        return None
    start_time = time.time()

    # region console logger initialization
    logger_handlers = []
    if logger is ...:
        logger = logging.getLogger(__name__)
        logger_handlers.append(gdalos_logger.set_logger_console(logger))
        logger.debug("console handler added")
    all_args["logger"] = logger
    verbose = logger is not None and logger is not ...
    if verbose:
        print_time_now(logger)
        logger.info(f'gdal version: {gdal.__version__}')
        logger.info(all_args)
    # endregion

    # creating a copy of the input dictionaries, as I don't want to change the input
    no_yes = ("NO", "YES")
    config_options = dict(config_options or dict())
    common_options = dict(common_options or dict())
    creation_options = dict(creation_options or dict())
    translate_options = dict(translate_options or dict())
    warp_options = dict(warp_options or dict())
    out_suffixes = []

    if ovr_idx not in [..., None]:
        ovr_idx = gdalos_util.get_ovr_idx(filename, ovr_idx)

    ds = gdalos_util.open_ds(filename, ovr_idx=ovr_idx, open_options=open_options, logger=logger)
    # region decide which overviews to make
    if ovr_idx is not ...:
        warp_options['overviewLevel'] = 'None'  # force gdal to use the selected ovr_idx
        if ovr_idx is None:
            ovr_idx = 0  # base raster
        overview_count = gdalos_util.get_ovr_count(ds)
        src_ovr_last = ovr_idx + overview_count
        if dst_ovr_count is not None:
            if dst_ovr_count >= 0:
                src_ovr_last = ovr_idx + min(overview_count, dst_ovr_count)
            else:
                # in this case we'll reopen the ds with a new ovr_idx, because we want only the last overviews
                new_src_ovr = max(0, overview_count + dst_ovr_count + 1)
                if new_src_ovr != ovr_idx:
                    ovr_idx = new_src_ovr
                    del ds
                    ds = gdalos_util.open_ds(filename, ovr_idx=ovr_idx, open_options=open_options, logger=logger)
                    overview_count = gdalos_util.get_ovr_count(ds)
                    src_ovr_last = ovr_idx + overview_count
    # endregion

    # filename_is_ds = not isinstance(filename, (str, Path))
    filename_is_ds = ds == filename

    try:
        gdal_version = int(gdal.VersionInfo())
        support_of_cog = gdal_version >= 3_01_00_00
        if verbose:
            logger.info(f'gdal version: {gdal_version}; cog driver support: {support_of_cog}')
    except:
        support_of_cog = False
    if cog is ...:
        cog = not filename_is_ds
    if ovr_type is None:
        cog = False
    if return_ds is ...:
        return_ds = of == GdalOutputFormat.mem
    if of is ... or of is None:
        of = GdalOutputFormat.mem if return_ds else GdalOutputFormat.cog if (
                cog and support_of_cog) else GdalOutputFormat.gtiff
    elif isinstance(of, str):
        of = of.lower()
        try:
            of = GdalOutputFormat[of]
        except:
            pass
    if of == GdalOutputFormat.cog:
        cog = True

    if of == GdalOutputFormat.mem:
        write_info = False
        write_spec = False
    if filename_is_ds:
        input_ext = None
    else:
        filename = Path(filename.strip())
        if os.path.isdir(filename):
            raise Exception('input is a dir, not a file: "{}"'.format(filename))
        if not os.path.isfile(filename):
            raise OSError('file not found: "{}"'.format(filename))
        input_ext = os.path.splitext(filename)[1].lower()

    if extent is None:
        extent_in_4326 = True

    if sparse_ok is ...:
        sparse_ok = not filename_is_ds and str(Path(filename).suffix).lower() == '.vrt'
    creation_options["SPARSE_OK"] = no_yes[bool(sparse_ok)]

    extent_was_cropped = False
    extent_aligned = True
    input_is_vrt = input_ext == ".vrt"

    geo_transform = ds.GetGeoTransform()
    input_res = (geo_transform[1], geo_transform[5])

    band_types = gdalos_util.get_band_types(ds)
    if kind in [None, ...]:
        kind = RasterKind.guess(band_types)

    # region warp CRS handling
    pjstr_src_srs = projdef.get_srs_pj_from_ds(ds)
    pjstr_tgt_srs = None
    tgt_zone = None
    if warp_srs is not None:
        pjstr_tgt_srs, tgt_zone1 = projdef.parse_proj_string_and_zone(warp_srs)
        tgt_datum, tgt_zone = projdef.get_datum_and_zone_from_projstring(pjstr_tgt_srs)
        if tgt_zone1:
            tgt_zone = tgt_zone1
        if projdef.proj_is_equivalent(pjstr_src_srs, pjstr_tgt_srs):
            warp_srs = None  # no warp is really needed here
    if warp_srs is not None:
        extent_aligned = False
        if tgt_zone is not None and extent_in_4326:
            if tgt_zone != 0:
                # cropping according to tgt_zone bounds
                zone_extent = GeoRectangle.from_points(projdef.get_utm_zone_extent_points(tgt_zone))
                if extent is None:
                    extent = zone_extent
                elif extent_crop_to_minimal:
                    extent = extent.intersect(zone_extent)
                extent_was_cropped = True
        out_suffixes.append(projdef.get_canonic_name(tgt_datum, tgt_zone))
        if kind == RasterKind.dtm:
            common_options["outputType"] = gdal.GDT_Float32  # 'Float32'
        warp_options["dstSRS"] = pjstr_tgt_srs
    # endregion

    if cutline:
        if isinstance(cutline, str):
            cutline_filename = cutline
        elif isinstance(cutline, Sequence):
            cutline_filename = tempfile.mktemp(suffix='.gpkg')
            temp_files.append(cutline_filename)
            gdalos_util.wkt_write_ogr(cutline_filename, cutline, of='GPKG')
        warp_options['cutlineDSName'] = cutline_filename

    do_warp = warp_srs is not None or cutline is not None

    # region compression
    resample_is_needed = warp_srs is not None or (out_res not in [..., None])
    org_comp = gdalos_util.get_image_structure_metadata(ds, "COMPRESSION")
    src_is_lossy = (org_comp is not None) and ("JPEG" in org_comp)
    if lossy in [None, ...]:
        lossy = src_is_lossy or resample_is_needed
    if lossy and (kind != RasterKind.photo):
        lossy = False
    if not lossy:
        comp = "DEFLATE"
    else:
        comp = "JPEG"
        out_suffixes.append("jpg")
    jpeg_compression = (comp == "JPEG") and (len(band_types) in (3, 4))
    # endregion

    # region expand_rgb
    if kind == RasterKind.pal and expand_rgb:
        translate_options["rgbExpand"] = "rgb"
        out_suffixes.append("rgb")
    # endregion

    # region nodatavalue
    if dst_nodatavalue is ...:
        if kind == RasterKind.dtm:
            dst_nodatavalue = default_multi_byte_nodata_value
        else:
            dst_nodatavalue = None  # don't change no_data
    if dst_nodatavalue is not None:
        src_nodatavalue_org = gdalos_util.get_nodatavalue(ds)
        if src_nodatavalue is ...:
            src_nodatavalue = src_nodatavalue_org
        elif src_nodatavalue is None:
            # assume raster minimum is nodata if nodata isn't set
            logger.debug('finding raster minimum, this might take some time...')
            src_min_value = gdalos_util.get_raster_minimum(ds)
            # if abs(src_min_value - default_multi_byte_nodata_value) < 100:
            # assuming that the raster minimum is indeed a nodatavalue if it's very low
            if True:
                src_nodatavalue = src_min_value
        if src_nodatavalue is not None:
            if src_nodatavalue != dst_nodatavalue:
                do_warp = True
                warp_options["dstNodata"] = dst_nodatavalue

            if src_nodatavalue_org != src_nodatavalue:
                translate_options["noData"] = src_nodatavalue
                warp_options["srcNodata"] = src_nodatavalue
    # endregion

    # region extent
    org_extent_in_src_srs = gdalos_extent.get_extent(ds)
    if org_extent_in_src_srs.is_empty():
        raise Exception(f"no input extent: {filename} [{org_extent_in_src_srs}]")
    out_extent_in_src_srs = org_extent_in_src_srs
    if extent is not None or partition is not None:
        pjstr_4326 = projdef.get_proj_string("w")  # 'EPSG:4326'
        transform = projdef.get_transform(pjstr_src_srs, pjstr_4326)
        src_srs_is_4326 = transform is None
        if extent is None:
            extent = gdalos_extent.translate_extent(org_extent_in_src_srs, transform)
        if pjstr_tgt_srs is None:
            pjstr_tgt_srs = pjstr_src_srs
            transform = None
        else:
            transform = projdef.get_transform(pjstr_src_srs, pjstr_tgt_srs)

        org_extent_in_tgt_srs = gdalos_extent.translate_extent(org_extent_in_src_srs, transform)
        if org_extent_in_tgt_srs.is_empty():
            raise Exception(f"no input extent: {filename} [{org_extent_in_tgt_srs}]")

        if extent_in_4326:
            if extent_crop_to_minimal and src_srs_is_4326:
                # we better intersect in 4326 then in a projected srs
                # because the projected srs bounds might be wider then the 'valid' zone bounds
                extent = extent.intersect(org_extent_in_src_srs)
            transform = projdef.get_transform(pjstr_4326, pjstr_tgt_srs)
            out_extent_in_tgt_srs = gdalos_extent.translate_extent(extent, transform)
        else:
            out_extent_in_tgt_srs = extent
        if extent_crop_to_minimal:
            out_extent_in_tgt_srs = out_extent_in_tgt_srs.intersect(org_extent_in_tgt_srs)

        if out_extent_in_tgt_srs != org_extent_in_tgt_srs:
            extent_was_cropped = True
            transform = projdef.get_transform(pjstr_tgt_srs, pjstr_4326)
            extent = gdalos_extent.translate_extent(out_extent_in_tgt_srs, transform)

        if out_extent_in_tgt_srs.is_empty():
            raise Exception(f"no output extent: {filename} [{out_extent_in_tgt_srs}]")

        if partition:
            out_extent_in_tgt_srs_part = out_extent_in_tgt_srs.get_partition(partition)
        else:
            out_extent_in_tgt_srs_part = out_extent_in_tgt_srs

        if extent_aligned:
            out_extent_in_tgt_srs_part = out_extent_in_tgt_srs_part.align(geo_transform)

        # -projwin minx maxy maxx miny (ulx uly lrx lry)
        translate_options["projWin"] = out_extent_in_tgt_srs_part.lurd
        # -te minx miny maxx maxy
        warp_options["outputBounds"] = out_extent_in_tgt_srs_part.ldru

        transform = projdef.get_transform(pjstr_4326, pjstr_src_srs)
        out_extent_in_src_srs = gdalos_extent.translate_extent(extent, transform)
        if extent_crop_to_minimal:
            out_extent_in_src_srs = out_extent_in_src_srs.intersect(org_extent_in_src_srs)
        if out_extent_in_src_srs.is_empty():
            raise Exception
    elif src_win is not None:
        translate_options["srcWin"] = src_win
    # endregion

    if do_warp and warp_error_threshold is not None:
        warp_options['errorThreshold'] = warp_error_threshold
    if do_warp and warp_scale:
        if isinstance(warp_scale, str):
            warp_scale = float(warp_scale)
        if not isinstance(warp_scale, Sequence):
            warp_scale = [warp_scale, warp_scale]
        warp_options["warpOptions"] = ['XSCALE={}'.format(warp_scale[0]), 'YSCALE={}'.format(warp_scale[1])]

    # region out_res
    if out_res is not ...:
        if out_res is not None:
            if isinstance(out_res, str):
                out_res = float(out_res)
            if not isinstance(out_res, Sequence):
                out_res = [out_res, -out_res]
        elif warp_srs is not None:
            transform_src_tgt = projdef.get_transform(pjstr_src_srs, pjstr_tgt_srs)
            if transform_src_tgt is not None:
                #  out_res is None and warp
                out_res = gdalos_extent.transform_resolution(transform_src_tgt, input_res, out_extent_in_src_srs)
        if out_res is not None:
            common_options["xRes"], common_options["yRes"] = out_res
            warp_options["targetAlignedPixels"] = True
            out_suffixes.append(str(out_res))
        elif ovr_idx is not ...:
            #  out_res is None and no warp
            out_res = input_res
    # endregion

    # region decide ovr_type
    trans_or_warp_is_needed = bool(
        do_warp
        or extent
        or src_win
        or resample_is_needed
        or (lossy != src_is_lossy)
        or translate_options
    )

    if ovr_type == OvrType.auto_select:
        if (ovr_idx is not ...) and (overview_count > 0) or (of == GdalOutputFormat.cog):
            # if the raster has overviews then use them, otherwise create overviews
            ovr_type = OvrType.existing_reuse
        else:
            ovr_type = OvrType.create_external_auto
    if (ovr_type == OvrType.existing_reuse) and (ovr_idx is ...):
        raise Exception('ovr_idx = {} cannot be used with ovr_type == {}'.format(..., OvrType.existing_reuse))

    cog_2_steps = \
        cog and \
        (trans_or_warp_is_needed or
         ovr_type in [OvrType.create_external_auto, OvrType.create_external_single,
                      OvrType.create_external_multi, OvrType.create_internal])

    if cog_2_steps:
        if of == GdalOutputFormat.cog:
            if prefer_2_step_cog is ...:
                prefer_2_step_cog = trans_or_warp_is_needed and (ovr_type == OvrType.existing_reuse)
            if prefer_2_step_cog:
                of = GdalOutputFormat.gtiff
            else:
                cog_2_steps = False
    # endregion

    # region make out_filename
    if outext in [..., None]:
        outext = gdalos_util.get_ext_by_of(of)
    elif not outext.startswith('.'):
        outext = '.' + outext
    auto_out_filename = out_filename is None and of != GdalOutputFormat.mem
    if auto_out_filename:
        if filename_is_ds:
            raise Exception('input is ds and no output filename is given')
        if (
                cog_2_steps
                and ovr_type == OvrType.create_external_auto
                and not trans_or_warp_is_needed
                and not input_is_vrt
        ):
            # create overviews for the input file then create a cog
            out_filename = filename
        else:
            out_extent_in_4326 = extent
            if extent_was_cropped and (out_extent_in_src_srs is not None):
                transform = projdef.get_transform(pjstr_src_srs, pjstr_4326)
                out_extent_in_4326 = gdalos_extent.translate_extent(
                    out_extent_in_src_srs, transform
                )
            if out_extent_in_4326 is not None:
                out_suffixes.append(
                    "x[{},{}]_y[{},{}]".format(
                        *(round(x, 2) for x in out_extent_in_4326.lrdu)
                    )
                )
            elif src_win is not None:
                out_suffixes.append("off[{},{}]_size[{},{}]".format(*src_win))
            if partition is not None:
                partition_str = "part_"
                if partition.w > 1:
                    partition_str += "x[{},{}]".format(partition.x, partition.w)
                if partition.h > 1:
                    partition_str += "y[{},{}]".format(partition.y, partition.h)
                out_suffixes.append(partition_str)
            if not out_suffixes:
                if outext.lower() == input_ext:
                    out_suffixes.append("new")
            if out_suffixes:
                out_suffixes = "." + ".".join(out_suffixes)
            else:
                out_suffixes = ""
            out_filename = gdalos_util.concat_paths(
                filename, out_suffixes + outext
            )
            if keep_src_ovr_suffixes and ovr_idx is not ...:
                out_filename = gdalos_util.concat_paths(
                    out_filename, ".ovr" * ovr_idx
                )
    elif out_filename:
        out_filename = Path(out_filename)
    else:
        out_filename = ''

    if out_path is not None:
        if out_path_with_src_folders:
            out_src_path = out_filename.parts[1:]
        else:
            out_src_path = [out_filename.parts[-1]]
        out_filename = Path(out_path).joinpath(*out_src_path)

    if out_filename and not os.path.exists(os.path.dirname(out_filename)):
        os.makedirs(os.path.dirname(out_filename), exist_ok=True)

    if cog:
        if auto_out_filename:
            final_filename = out_filename.with_suffix(".cog" + outext)
            if not cog_2_steps:
                out_filename = final_filename
        else:
            final_filename = out_filename
            if cog_2_steps:
                out_filename = out_filename.with_suffix(".temp" + outext)
    else:
        final_filename = out_filename
    # endregion

    if cog_2_steps:
        final_files_for_step_1 = temp_files
        ovr_files_for_step_1 = temp_files
    else:
        final_files_for_step_1 = final_files
        ovr_files_for_step_1 = ovr_files

    final_output_exists = do_skip_if_exists(final_filename, overwrite, logger)

    # for OvrType.existing_reuse we'll create the files in backwards order in the overview creation step
    skipped = (final_output_exists or
               ((not filename_is_ds) and
                ((ovr_type == OvrType.existing_reuse and (not cog or cog_2_steps)) or
                 ((filename == out_filename) and (not trans_or_warp_is_needed)))))

    if final_output_exists:
        final_files.append(final_filename)
    elif write_spec:
        spec_filename = gdalos_util.concat_paths(out_filename, ".spec")
        logger_handlers.append(gdalos_logger.set_file_logger(logger, spec_filename))
        logger.debug('spec file handler added: "{}"'.format(spec_filename))
        logger.debug("gdalos versoin: {}".format(__version__))
        aux_files.append(spec_filename)
        # logger.debug('debug')
        # logger.info('info')
        # logger.warning('warning')
        # logger.error('error')
        # logger.critical('critical')

    # region create base raster
    ret_code = None
    out_ds = None
    if not skipped:
        # region jpeg
        if jpeg_compression:
            if of == GdalOutputFormat.gpkg:
                creation_options["TILE_FORMAT"] = "JPEG"
            elif of != GdalOutputFormat.cog:
                creation_options["PHOTOMETRIC"] = "YCBCR"
            if quality and (quality is not ...):
                creation_options["JPEG_QUALITY" if of == GdalOutputFormat.gtiff else 'QUALITY'] = str(quality)

            # alpha channel is not supported with PHOTOMETRIC=YCBCR, thus we drop it or keep it as a mask band
            if len(band_types) == 4:
                if do_warp:
                    raise Exception(
                        "this mode is not supported: warp RGBA raster with JPEG output. "
                        "You could do it in two steps: "
                        "1. warp with lossless output, "
                        "2. save as jpeg"
                    )
                else:
                    translate_options["bandList"] = [1, 2, 3]
                    if keep_alpha is ...:
                        keep_alpha = filename_is_ds or input_ext != '.gpkg'
                    if keep_alpha:
                        translate_options["maskBand"] = 4  # keep the alpha band as mask
        # endregion

        common_options["format"] = enum_to_str(of)
        if of in [GdalOutputFormat.gtiff, GdalOutputFormat.cog]:
            if not big_tiff:
                big_tiff = False
            elif big_tiff is ...:
                big_tiff = "IF_SAFER"
            elif not isinstance(big_tiff, str):
                big_tiff = no_yes[bool(big_tiff)]

            creation_options["BIGTIFF"] = big_tiff
            creation_options["COMPRESS"] = comp

        tiled = tiled.upper() != no_yes[False] if isinstance(tiled, str) else bool(tiled)
        if of == GdalOutputFormat.gtiff:
            creation_options["TILED"] = no_yes[tiled]
        if tiled and block_size is not ...:
            # if block_size is ...:
            #     block_size = 256  # default gdal block_size
            # assert(width / BlockXSize * height / BlockYSize * (big_tiff ? 8: 4) <= 0x80000000)
            # File too large regarding tile size. This would result
            # in a file with tile arrays larger than 2GB
            if of == GdalOutputFormat.gtiff:
                creation_options["BLOCKXSIZE"] = block_size
                creation_options["BLOCKYSIZE"] = block_size
            elif of == GdalOutputFormat.cog:
                creation_options["BLOCKSIZE"] = block_size
        if cog and not cog_2_steps and of != GdalOutputFormat.cog:
            creation_options["COPY_SRC_OVERVIEWS"] = "YES"

        if creation_options:
            creation_options_list = []
            for k, v in creation_options.items():
                creation_options_list.append("{}={}".format(k, v))
            common_options["creationOptions"] = creation_options_list

        if print_progress:
            common_options["callback"] = print_progress_callback(print_progress)

        if resample_is_needed:
            if resampling_alg is ...:
                resampling_alg = resampling_alg_by_kind(kind, expand_rgb)
            if resampling_alg is not None:
                common_options["resampleAlg"] = enum_to_str(resampling_alg)

        if verbose:
            logger.info('filename: "' + str(out_filename) + '" ...')
            if common_options:
                logger.info("common options: " + str(common_options))

        if input_ext == ".xml":
            config_options = {"GDAL_HTTP_UNSAFESSL": "YES"}  # for gdal-wms xml files

        try:
            if config_options:
                if verbose:
                    logger.info("config options: " + str(config_options))
                for k, v in config_options.items():
                    gdal.SetConfigOption(k, v)

            if do_warp:
                if verbose and warp_options:
                    logger.info("warp options: " + str(warp_options))
                out_ds = gdal.Warp(str(out_filename), ds, **common_options, **warp_options)
            else:
                if verbose and translate_options:
                    logger.info("translate options: " + str(translate_options))
                out_ds = gdal.Translate(str(out_filename), ds, **common_options, **translate_options)
            ret_code = out_ds is not None
            if not return_ds:
                out_ds = None  # close output ds
            if ret_code:
                final_files_for_step_1.append(out_filename)
        except Exception as e:
            if verbose:
                logger.error(str(e))
        finally:
            for key, val in config_options.items():
                gdal.SetConfigOption(key, None)

        if verbose:
            seconds = round(time.time() - start_time)
            time_string = str(datetime.timedelta(seconds=seconds))
            logger.info(
                'time: {} for creating file: "{}"'.format(time_string, out_filename)
            )

    if verbose:
        logger.debug('closing file: "{}"'.format(filename))
    ds = None  # close input ds if filename was input
    # end region

    # region create overviews, cog, info
    cog_ready = cog and final_output_exists
    if not cog_ready and (ret_code or skipped):
        if not skipped and hide_nodatavalue:
            gdalos_util.unset_nodatavalue(out_ds or str(out_filename))

        if not cog or cog_2_steps:
            # create overview file(s)
            if ovr_type == OvrType.existing_reuse:
                # overviews are numbered as follows (i.e. for dst_ovr_count=3, meaning create base+3 ovrs=4 files):
                # -1: base dataset, 0: first ovr, 1: second ovr, 2: third ovr

                all_args_new = all_args.copy()
                all_args_new["ovr_type"] = None
                all_args_new["dst_ovr_count"] = None
                all_args_new["out_path"] = None
                all_args_new["write_spec"] = False
                all_args_new["cog"] = False
                all_args_new["logger"] = logger
                # iterate backwards on the overviews
                if verbose:
                    logger.debug(
                        "iterate on overviews creation, from {} to {}".format(
                            src_ovr_last, ovr_idx
                        )
                    )
                for cur_ovr_idx in range(src_ovr_last, ovr_idx - 1, -1):
                    all_args_new["final_files"] = []
                    all_args_new["ovr_files"] = []  # there shouldn't be any
                    all_args_new["aux_files"] = []
                    all_args_new["temp_files"] = []  # there shouldn't be any
                    all_args_new["out_filename"] = gdalos_util.concat_paths(
                        out_filename, ".ovr" * (cur_ovr_idx - ovr_idx)
                    )
                    all_args_new["ovr_idx"] = cur_ovr_idx
                    if out_res not in [None, ...]:
                        res_factor = 2 ** (cur_ovr_idx - ovr_idx)
                        all_args_new["out_res"] = [r * res_factor for r in out_res]
                    all_args_new["write_info"] = (
                            write_info and (cur_ovr_idx == ovr_idx) and not cog
                    )
                    ret_code = gdalos_trans(**all_args_new)
                    if not ret_code:
                        logger.warning(
                            "return code was None for creating {}".format(
                                all_args_new["out_filename"]
                            )
                        )
                    if cur_ovr_idx == ovr_idx:
                        final_files_for_step_1.extend(all_args_new["final_files"])
                    else:
                        ovr_files_for_step_1.extend(all_args_new["final_files"])
                    if verbose:
                        for f in ["ovr_files", "temp_files"]:
                            if all_args_new[f]:
                                logger.error(
                                    "there shound not be any {} here, but there are! {}".format(
                                        f, all_args_new[f]
                                    )
                                )
                        if len(all_args_new["final_files"]) != 1:
                            logger.error(
                                "ovr creating should have made exactly 1 file! {}".format(
                                    all_args_new["final_files"]
                                )
                            )
                    aux_files.extend(all_args_new["aux_files"])
                write_info = write_info and cog
            elif not filename_is_ds and ovr_type not in [None, OvrType.existing_reuse]:
                # create overviews from dataset (internal or external)
                ret_code = gdalos_ovr(
                    out_filename,
                    overwrite=overwrite,
                    ovr_type=ovr_type,
                    dst_ovr_count=dst_ovr_count,
                    kind=kind,
                    resampling_alg=resampling_alg,
                    print_progress=print_progress,
                    logger=logger,
                    ovr_files=ovr_files_for_step_1,
                )

        if cog_2_steps:
            if verbose:
                logger.debug("running cog 2nd step...")
            cog_final_files = []  # there should be exactly one!
            cog_temp_files = []  # there shouldn't be any!
            cog_ovr_files = []  # there shouldn't be any!
            cog_aux_files = []
            ret_code = gdalos_trans(
                out_filename,
                out_filename=final_filename,
                cog=True,
                ovr_type=OvrType.existing_reuse,
                of=of,
                outext=outext,
                tiled=tiled,
                big_tiff=big_tiff,
                print_progress=print_progress,
                final_files=cog_final_files,
                ovr_files=cog_ovr_files,
                aux_files=cog_aux_files,
                temp_files=cog_temp_files,
                delete_temp_files=False,
                logger=logger,
                overwrite=overwrite,
                write_info=write_info,
                write_spec=False,
            )
            if ret_code and verbose:
                if cog_temp_files:
                    logger.error(
                        "cog 2nd step should not have any temp files, but it has! {}".format(
                            cog_temp_files
                        )
                    )
                if len(cog_final_files) != 1:
                    logger.error(
                        "cog 2nd step should have made exactly 1 file! {}".format(
                            cog_final_files
                        )
                    )
            final_files.extend(cog_final_files)
            aux_files.extend(cog_aux_files)
            write_info = (
                False  # we don't need an info for the temp file from first step
            )
        if write_info:
            info = gdalos_info(
                out_filename, overwrite=overwrite, logger=logger
            )
            if info is not None:
                aux_files.append(info)
    # endregion

    # region log file lists
    if verbose:
        if final_files:
            logger.info("final_files: {}".format(final_files))
        if aux_files:
            logger.info("aux_files: {}".format(aux_files))
        if temp_files:
            if delete_temp_files:
                deleted = ""
            else:
                deleted = "not "
            logger.info(
                "temp_files (will {}be deleted): {}".format(deleted, temp_files)
            )
    # endregion

    # region delete temp files
    if delete_temp_files and temp_files:
        for f in temp_files:
            if f == filename:
                if verbose:
                    logger.error(
                        'somehow the input file was set as a temp file for deletion: "{}")'.format(
                            f
                        )
                    )
            elif os.path.isfile(f):
                try:
                    os.remove(f)
                except Exception as e:
                    logger.warning('could not delete file: "{}" ({})'.format(f, str(e)))
            else:
                logger.warning('file for deletion not found: "{}"'.format(f))
        temp_files.clear()
    # endregion

    if verbose:
        logger.info("*** done! ***\n")

    # region remove loggers
    if logger_handlers:
        logger.debug("removing {} logging handlers".format(len(logger_handlers)))
        for handler in logger_handlers:
            handler.close()
            logger.removeHandler(handler)
        logger.debug("logging handlers removed")  # this shouldn't log anything
    # end region
    return out_ds or ret_code


def add_ovr(
        filename,
        options,
        access_mode,
        overwrite=True,
        logger=None,
        ovr_files: list = None,
):
    verbose = logger is not None and logger is not ...
    filename = Path(filename)
    out_filename = gdalos_util.concat_paths(filename, ".ovr")
    if not do_skip_if_exists(out_filename, overwrite, logger):
        if verbose:
            logger.info(
                'adding ovr for: "{}" options: {} access_mode: {}'.format(
                    out_filename, options, access_mode
                )
            )
        with gdalos_util.OpenDS(filename, access_mode=access_mode, logger=logger) as ds:
            ret_code = ds.BuildOverviews(**options) == 0
            if ret_code and ovr_files is not None:
                ovr_files.append(out_filename)
            return ret_code
    else:
        return None


default_dst_ovr_count = 10


def gdalos_ovr(
        filename,
        comp=None,
        overwrite=True,
        ovr_type=...,
        dst_ovr_count=default_dst_ovr_count,
        kind=None,
        resampling_alg=...,
        config_options: dict = None,
        ovr_options: dict = None,
        ovr_files: list = None,
        print_progress=...,
        logger=None,
):
    verbose = logger is not None and logger is not ...
    filename = Path(filename)
    if os.path.isdir(filename):
        raise Exception('input is a dir, not a file: "{}"'.format(filename))

    if not os.path.isfile(filename):
        raise Exception('file not found: "{}"'.format(filename))

    if dst_ovr_count is None or dst_ovr_count <= 0:
        dst_ovr_count = default_dst_ovr_count

    if ovr_files is None:
        ovr_files = []

    if isinstance(ovr_type, str):
        ovr_type = OvrType[ovr_type]
    elif ovr_type is ...:
        ovr_type = OvrType.auto_select
    if ovr_type in [OvrType.auto_select, OvrType.create_external_auto]:
        file_size = os.path.getsize(filename)
        max_ovr_gb = 1
        if file_size > max_ovr_gb * 1024 ** 3:
            ovr_type = OvrType.create_external_multi
        else:
            ovr_type = OvrType.create_external_single
    elif ovr_type not in [
        OvrType.create_internal,
        OvrType.create_external_single,
        OvrType.create_external_multi,
    ]:
        return None

    if ovr_options is None:
        ovr_options = dict()
    if config_options is None:
        config_options = dict()
    if resampling_alg is ...:
        if kind in [None, ...]:
            kind = RasterKind.guess(filename)
        resampling_alg = resampling_alg_by_kind(kind)
    if resampling_alg is not None:
        ovr_options["resampling"] = enum_to_str(resampling_alg)
    if print_progress:
        ovr_options["callback"] = print_progress_callback(print_progress)

    if comp is None:
        comp = gdalos_util.get_image_structure_metadata(filename, "COMPRESSION")
    if comp == "YCbCr JPEG":
        config_options["COMPRESS_OVERVIEW"] = "JPEG"
        config_options["PHOTOMETRIC_OVERVIEW"] = "YCBCR"
        config_options["INTERLEAVE_OVERVIEW"] = "PIXEL"
    else:
        config_options["COMPRESS_OVERVIEW"] = comp

    try:
        if config_options:
            if verbose:
                logger.info("config options: " + str(config_options))
            for k, v in config_options.items():
                gdal.SetConfigOption(k, v)

        out_filename = filename
        access_mode = gdal.GA_ReadOnly
        if ovr_type in (OvrType.create_internal, OvrType.create_external_single):
            if ovr_type == OvrType.create_internal:
                access_mode = gdal.GA_Update
            ovr_levels = []
            for i in range(dst_ovr_count):
                ovr_levels.append(
                    2 ** (i + 1)
                )  # ovr_levels = '2 4 8 16 32 64 128 256 512 1024'
            ovr_options["overviewlist"] = ovr_levels
            ret_code = add_ovr(
                out_filename,
                ovr_options,
                access_mode,
                overwrite,
                logger,
                ovr_files,
            )
        elif ovr_type == OvrType.create_external_multi:
            ovr_options["overviewlist"] = [2]
            ret_code = None
            for i in range(dst_ovr_count):
                ret_code = add_ovr(
                    filename,
                    ovr_options,
                    access_mode,
                    overwrite,
                    logger,
                    ovr_files,
                )
                if not ret_code:
                    break
                filename = gdalos_util.concat_paths(filename, ".ovr")
        else:
            raise Exception("invalid ovr type")
    finally:
        for key, val in config_options.items():
            gdal.SetConfigOption(key, None)
    return ret_code


def gdalos_info(filename_or_ds, overwrite=True, logger=None):
    filename_or_ds = Path(filename_or_ds)
    if os.path.isdir(filename_or_ds):
        raise Exception(f"input is a dir, not a file: {filename_or_ds}")
    if not os.path.isfile(filename_or_ds):
        raise Exception("file not found: {}".format(filename_or_ds))
    out_filename = gdalos_util.concat_paths(filename_or_ds, ".info")
    if not do_skip_if_exists(out_filename, overwrite=overwrite):
        with gdalos_util.OpenDS(filename_or_ds, logger=logger) as ds:
            gdal_info = gdal.Info(ds)
        with open(out_filename, "w") as w:
            w.write(gdal_info)
        return out_filename
    else:
        return None


def gdalos_vrt(
        filenames: MaybeSequence,
        filenames_expand: Union[type(...), bool] = ...,
        vrt_path=None,
        kind=None,
        resampling_alg=...,
        overwrite=True,
        logger=None,
):
    if gdalos_util.is_list_like(filenames):
        flatten_filenames = gdalos_util.flatten_and_expand_file_list(filenames, do_expand_glob=filenames_expand)
    else:
        flatten_filenames = [filenames]
    flatten_filenames = [str(f) for f in flatten_filenames]
    if not flatten_filenames:
        return None
    first_filename = flatten_filenames[0]
    if vrt_path is None:
        vrt_path = first_filename + ".vrt"

    if os.path.isdir(vrt_path):
        vrt_path = os.path.join(vrt_path, os.path.basename(first_filename) + ".vrt")
    if do_skip_if_exists(vrt_path, overwrite, logger):
        return vrt_path
    if os.path.isfile(vrt_path):
        raise Exception("could not delete vrt file: {}".format(vrt_path))
    os.makedirs(os.path.dirname(vrt_path), exist_ok=True)
    vrt_options = dict()
    if resampling_alg is ...:
        if kind in [None, ...]:
            kind = RasterKind.guess(first_filename)
        resampling_alg = resampling_alg_by_kind(kind)
    if resampling_alg is not None:
        vrt_options["resampleAlg"] = resampling_alg
    vrt_options = gdal.BuildVRTOptions(*vrt_options)
    out_ds = gdal.BuildVRT(vrt_path, flatten_filenames, options=vrt_options)
    if out_ds is None:
        return None
    del out_ds
    success = os.path.isfile(vrt_path)
    if success:
        return vrt_path
    else:
        return None



from setuptools import setup, find_packages

from src.gdalos_data.__data__ import (
    __author__,
    __author_email__,
    __license__,
    __url__,
    __version__,
)

setup(
    name="gdalos",
    version=__version__,
    author=__author__,
    author_email=__author_email__,
    license=__license__,
    url=__url__,
    description="a simple gdal translate/warp/addo python wrapper for raster batch processing",
    # packages=["gdalos", "gdalos_data", "gdalos_qt"],
    packages=find_packages("src"),  # include all packages under src
    package_dir={"": "src"},   # tell distutils packages are under src
    install_requires=["gdal"],
    extras_require={"PyQt": ["fidget", "PyQt5"], "PySide": ["fidget", "PySide2"]},
    python_requires=">=3.6.0",
    include_package_data=True,
    data_files=[("", ["README.rst", "LICENSE"])],
)



from gdalos.talos.util import NormalizeAngleDeg
from gdalos.talos.gen_consts import M_PI_180, M_2PI


def GetFromToAngle(DirectionDeg, ApertureDeg: float) -> (float, float):
    if ApertureDeg >= 360 - 1e-10:
        AFromRad = 0
        AToRad = 0
    else:
        HeadDir = NormalizeAngleDeg(90 - DirectionDeg)
        AFromRad = NormalizeAngleDeg(HeadDir - ApertureDeg * 0.5) * M_PI_180
        AToRad = NormalizeAngleDeg(HeadDir + ApertureDeg * 0.5) * M_PI_180
        if AFromRad > AToRad:
            AFromRad = AFromRad - M_2PI
    return AFromRad, AToRad

